# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from graphviz import Digraph

from ._provider import Provider, ProviderKind
from .typing import Graph, Item, Key, get_optional


@dataclass
class Node:
    name: str
    collapsed: bool = False


@dataclass
class FormattedProvider:
    name: str
    args: list[Node]
    ret: Node
    kind: ProviderKind


FormattedGraph = Dict[str, FormattedProvider]


def to_graphviz(
    graph: Graph,
    compact: bool = False,
    cluster_generics: bool = True,
    cluster_color: Optional[str] = '#f0f0ff',
    **kwargs: Any,
) -> Digraph:
    """
    Convert output of :py:class:`sciline.Pipeline.get_graph` to a graphviz graph.

    Parameters
    ----------
    graph:
        Output of :py:class:`sciline.Pipeline.get_graph`.
    compact:
        If True, parameter-table-dependent branches are collapsed into a single copy
        of the branch. Recommended for large graphs with long parameter tables.
    cluster_generics:
        If True, generic products are grouped into clusters.
    cluster_color:
        Background color of clusters. If None, clusters are dotted.
    kwargs:
        Keyword arguments passed to :py:class:`graphviz.Digraph`.
    """
    dot = Digraph(strict=True, **kwargs)
    formatted_graph = _format_graph(graph, compact=compact)
    ordered_graph = dict(
        sorted(formatted_graph.items(), key=lambda item: item[1].ret.name)
    )
    subgraphs = _to_subgraphs(ordered_graph)

    for origin, subgraph in subgraphs.items():
        cluster = cluster_generics and len(subgraph) > 1
        name = f'cluster_{origin}' if cluster else None
        with dot.subgraph(name=name) as dot_subgraph:
            if cluster:
                dot_subgraph.attr(rank='same')
                if cluster_color is None:
                    dot_subgraph.attr(style='dotted')
                else:
                    dot_subgraph.attr(style='filled', color=cluster_color)
            _add_subgraph(subgraph, dot, dot_subgraph)
    return dot


def _to_subgraphs(graph: FormattedGraph) -> Dict[str, FormattedGraph]:
    def get_subgraph_name(name: str) -> str:
        sgname = name.split('[')[0]
        if sgname == 'Series':
            # Example: Series[RowId, Material[Country]] -> RowId, Material[Country]
            return name.partition('[')[-1].rpartition(']')[0]
        return sgname

    subgraphs: Dict[str, FormattedGraph] = {}
    for p, formatted_p in graph.items():
        subgraph_name = get_subgraph_name(formatted_p.ret.name)
        if subgraph_name not in subgraphs:
            subgraphs[subgraph_name] = {}
        subgraphs[subgraph_name][p] = formatted_p
    return subgraphs


def _add_subgraph(graph: FormattedGraph, dot: Digraph, subgraph: Digraph) -> None:
    for p, formatted_p in graph.items():
        if formatted_p.kind == 'unsatisfied':
            subgraph.node(
                formatted_p.ret.name,
                formatted_p.ret.name,
                shape='box3d' if formatted_p.ret.collapsed else 'rectangle',
                color='red',
                fontcolor='red',  # Set text color to red
                style='dashed',
            )
        else:
            subgraph.node(
                formatted_p.ret.name,
                formatted_p.ret.name,
                shape='box3d' if formatted_p.ret.collapsed else 'rectangle',
            )
        # Do not draw the internal provider gathering index-dependent results into
        # a dict
        if formatted_p.kind == 'series':
            for arg in formatted_p.args:
                dot.edge(arg.name, formatted_p.ret.name, style='dashed')
        elif formatted_p.kind == 'function':
            dot.node(p, formatted_p.name, shape='ellipse')
            for arg in formatted_p.args:
                dot.edge(arg.name, p)
            dot.edge(p, formatted_p.ret.name)
        # else: Do not draw dummy providers created by Pipeline when setting instances


def _qualname(obj: Any) -> Any:
    return (
        obj.__qualname__ if hasattr(obj, '__qualname__') else obj.__class__.__qualname__
    )


def _format_graph(graph: Graph, compact: bool) -> FormattedGraph:
    return {
        _format_provider(provider, ret, compact=compact): FormattedProvider(
            name=_qualname(provider.func),
            args=[_format_type(a, compact=compact) for a in provider.arg_spec.keys()],
            ret=_format_type(ret, compact=compact),
            kind=provider.kind,
        )
        for ret, provider in graph.items()
    }


def _format_provider(provider: Provider, ret: Key, compact: bool) -> str:
    return f'{provider.location.qualname}_{_format_type(ret, compact=compact).name}'


T = TypeVar('T')


def _extract_type_and_labels(
    key: Union[Item[T], Type[T]], compact: bool
) -> Tuple[Type[T], List[Union[type, Tuple[type, Any]]]]:
    if isinstance(key, Item):
        label = key.label
        return key.tp, [lb.tp if compact else (lb.tp, lb.index) for lb in label]
    return key, []


def _format_type(tp: Key, compact: bool = False) -> Node:
    """
    Helper for _format_graph.

    If tp is a generic such as Array[float], we want to return 'Array[float]',
    but strip all module prefixes from the type name as well as the params.
    We may make this configurable in the future.
    """

    tp, labels = _extract_type_and_labels(tp, compact=compact)

    if (tp_ := get_optional(tp)) is not None:
        tp = tp_

    def get_base(tp: Key) -> str:
        return tp.__name__ if hasattr(tp, '__name__') else str(tp).split('.')[-1]

    def format_label(label: Union[type, Tuple[type, Any]]) -> str:
        if isinstance(label, tuple):
            tp, index = label
            return f'{get_base(tp)}={index}'
        return get_base(label)

    def with_labels(base: str) -> Node:
        if labels:
            return Node(
                name=f'{base}({", ".join([format_label(l) for l in labels])})',
                collapsed=compact,
            )
        return Node(name=base)

    if (origin := get_origin(tp)) is not None:
        params = [_format_type(param).name for param in get_args(tp)]
        return with_labels(f'{get_base(origin)}[{", ".join(params)}]')
    else:
        return with_labels(get_base(tp))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from dataclasses import dataclass
from typing import (
    Any,
    Generic,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from ._provider import Provider


@dataclass(frozen=True)
class Label:
    tp: type
    index: Any


T = TypeVar('T')


@dataclass(frozen=True)
class Item(Generic[T]):
    label: Tuple[Label, ...]
    tp: Type[T]


Key = Union[type, Item[Any]]
Graph = dict[Key, Provider]


def get_optional(tp: Key) -> Optional[Any]:
    if get_origin(tp) != Union:
        return None
    args = get_args(tp)
    if len(args) != 2 or type(None) not in args:
        return None
    return args[0] if args[1] == type(None) else args[1]  # noqa: E721


def get_union(tp: Key) -> Optional[Tuple[Any, ...]]:
    if get_origin(tp) != Union:
        return None
    return get_args(tp)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Optional

import pytest

from sciline.scheduler import DaskScheduler, NaiveScheduler, Scheduler


@pytest.fixture
def naive_scheduler() -> NaiveScheduler:
    return NaiveScheduler()


@pytest.fixture
def dask_scheduler() -> Optional[DaskScheduler]:
    try:
        import dask  # noqa: F401

        return DaskScheduler()
    except ModuleNotFoundError:
        return None


@pytest.fixture(params=['naive', 'dask'])
def scheduler(request: pytest.FixtureRequest) -> Scheduler:
    if request.param == 'naive':
        return request.getfixturevalue('naive_scheduler')  # type: ignore[no-any-return]

    sched = request.getfixturevalue('dask_scheduler')
    if sched is None:
        pytest.skip("Test requires Dask")
    return sched  # type: ignore[no-any-return]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""Handling of providers and their arguments."""
from __future__ import annotations

import inspect
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Literal,
    Optional,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

if TYPE_CHECKING:
    from .typing import Key


ToProvider = Callable[..., Any]
"""Callable that can be converted to a provider."""

ProviderKind = Literal[
    'function', 'parameter', 'series', 'table_cell', 'sentinel', 'unsatisfied'
]
"""Identifies the kind of a provider, most are used internally."""


class UnboundTypeVar(Exception):
    """
    Raised when a parameter of a generic provider is not bound to a concrete type.
    """


class Provider:
    """A provider.

    This class wraps a function that returns the provided values.
    That function can be a user-provided callable, in which case
    ``kind = 'function'``, or an internally constructed function
    for providing parameters or other special values.
    """

    def __init__(
        self,
        *,
        func: ToProvider,
        arg_spec: ArgSpec,
        kind: ProviderKind,
        location: Optional[ProviderLocation] = None,
    ) -> None:
        self._func = func
        self._arg_spec = arg_spec
        self._kind = kind
        self._location = (
            location if location is not None else ProviderLocation.from_function(func)
        )

    @classmethod
    def from_function(cls, func: ToProvider) -> Provider:
        """Construct from a function or other callable."""
        return cls(func=func, arg_spec=ArgSpec.from_function(func), kind='function')

    @classmethod
    def parameter(cls, param: Any) -> Provider:
        """Construct a provider that always returns the given value."""
        return cls(
            func=lambda: param,
            arg_spec=ArgSpec.null(),
            kind='parameter',
            location=ProviderLocation(
                name=f'param({type(param).__name__})', module=_module_name(param)
            ),
        )

    @classmethod
    def table_cell(cls, param: Any) -> Provider:
        """Construct a provider that returns the label for a table row."""
        return cls(
            func=lambda: param,
            arg_spec=ArgSpec.null(),
            kind='table_cell',
            location=ProviderLocation(
                name=f'table_cell({type(param).__name__})', module=_module_name(param)
            ),
        )

    @classmethod
    def provide_none(cls) -> Provider:
        """Provider that takes no arguments and returns None."""
        return cls(
            func=lambda: None,
            arg_spec=ArgSpec.null(),
            kind='function',
            location=ProviderLocation(name='provide_none', module='sciline'),
        )

    @property
    def func(self) -> ToProvider:
        """Return the function that implements the provider."""
        return self._func

    @property
    def arg_spec(self) -> ArgSpec:
        """Return the argument specification for the provider."""
        return self._arg_spec

    @property
    def kind(self) -> ProviderKind:
        """Return the kind of the provider."""
        return self._kind

    @property
    def location(self) -> ProviderLocation:
        """Return the location of the provider in source code."""
        return self._location

    def deduce_key(self) -> Any:
        """Attempt to determine the key (return type) of the provider."""
        if (key := get_type_hints(self._func).get('return')) is None:
            raise ValueError(
                f'Provider {self} lacks type-hint for return value or returns None.'
            )
        return key

    def bind_type_vars(self, bound: dict[TypeVar, Key]) -> Provider:
        """Replace TypeVars with their corresponding keys."""
        return Provider(
            func=self._func,
            arg_spec=self._arg_spec.bind_type_vars(bound),
            kind=self._kind,
        )

    def map_arg_keys(self, transform: Callable[[Key], Key]) -> Provider:
        """Return a new provider with transformed argument keys."""
        return Provider(
            func=self._func,
            arg_spec=self._arg_spec.map_keys(transform),
            kind=self._kind,
        )

    def __str__(self) -> str:
        return f"Provider('{self.location.name}')"

    def __repr__(self) -> str:
        return (
            f"Provider('{self.location.module}.{self.location.name}', "
            f"func={self._func})"
        )

    def call(self, values: dict[Key, Any]) -> Any:
        """Call the provider with arguments extracted from ``values``."""
        return self._func(
            *(values[arg] for arg in self._arg_spec.args),
            **{key: values[arg] for key, arg in self._arg_spec.kwargs},
        )


class ArgSpec:
    """Argument specification for a provider."""

    def __init__(self, *, args: dict[str, Key], kwargs: dict[str, Key]) -> None:
        """Build from components, use dedicated creation functions instead."""
        self._args = args
        self._kwargs = kwargs

    @classmethod
    def from_function(cls, provider: ToProvider) -> ArgSpec:
        """Parse the argument spec of a provider."""
        hints = get_type_hints(provider)
        signature = _get_provider_args(provider)
        try:
            args = {name: hints[name] for name in signature.args}
            kwargs = {name: hints[name] for name in signature.kwonlyargs}
        except KeyError:
            raise ValueError(
                f'Provider {provider} lacks type-hint for arguments.'
            ) from None
        return cls(args=args, kwargs=kwargs)

    @classmethod
    def from_args(cls, *args: Key) -> ArgSpec:
        """Create ArgSpec from positional arguments."""
        return cls(args={f'unknown_{i}': arg for i, arg in enumerate(args)}, kwargs={})

    @classmethod
    def null(cls) -> ArgSpec:
        """Create ArgSpec for a nullary function (no args)."""
        return cls(args={}, kwargs={})

    @property
    def args(self) -> Generator[Key, None, None]:
        yield from self._args.values()

    @property
    def kwargs(self) -> Generator[tuple[str, Key], None, None]:
        yield from self._kwargs.items()

    def keys(self) -> Generator[Key, None, None]:
        """Flat iterator over all argument types."""
        yield from self._args.values()
        yield from self._kwargs.values()

    def bind_type_vars(self, bound: dict[TypeVar, Key]) -> ArgSpec:
        """Bind concrete types to TypeVars."""
        return self.map_keys(lambda arg: _bind_free_typevars(arg, bound=bound))

    def map_keys(self, transform: Callable[[Key], Key]) -> ArgSpec:
        """Return a new ArgSpec with the keys mapped by ``callback``."""
        return ArgSpec(
            args={name: transform(arg) for name, arg in self._args.items()},
            kwargs={name: transform(arg) for name, arg in self._kwargs.items()},
        )


@dataclass
class ProviderLocation:
    name: str
    module: str

    @classmethod
    def from_function(cls, func: ToProvider) -> ProviderLocation:
        return cls(name=func.__name__, module=_module_name(func))

    @property
    def qualname(self) -> str:
        """Fully qualified name of the provider.

        Note that this always includes the module name unlike
        ``provider.func.__qualname__`` which depends on how the provider was imported.
        """
        if self.module:
            return f'{self.module}.{self.name}'
        return self.name


def _bind_free_typevars(tp: Union[TypeVar, Key], bound: dict[TypeVar, Key]) -> Key:
    if isinstance(tp, TypeVar):
        if (result := bound.get(tp)) is None:
            raise UnboundTypeVar(f'Unbound type variable {tp}')
        return result
    elif (origin := get_origin(tp)) is not None:
        result = origin[tuple(_bind_free_typevars(arg, bound) for arg in get_args(tp))]
        if result is None:
            raise ValueError(f'Binding type variables in {tp} resulted in `None`')
        return result
    else:
        return tp


def _module_name(x: Any) -> str:
    # getmodule might return None
    return getattr(inspect.getmodule(x), '__name__', '')


def _get_provider_args(func: ToProvider) -> inspect.FullArgSpec:
    if (wrapped := getattr(func, '__wrapped__', None)) is not None:
        return _get_provider_args(wrapped)
    return inspect.getfullargspec(func)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterable
from itertools import chain
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

from sciline.task_graph import TaskGraph

from ._provider import ArgSpec, Provider, ProviderLocation, ToProvider
from .display import pipeline_html_repr
from .domain import Scope, ScopeTwoParams
from .handler import (
    ErrorHandler,
    HandleAsBuildTimeException,
    HandleAsComputeTimeException,
    UnsatisfiedRequirement,
)
from .param_table import ParamTable
from .scheduler import Scheduler
from .series import Series
from .typing import Graph, Item, Key, Label, get_optional, get_union
from .utils import keyname

T = TypeVar('T')
KeyType = TypeVar('KeyType', bound=Key)
ValueType = TypeVar('ValueType', bound=Key)
IndexType = TypeVar('IndexType', bound=Key)
LabelType = TypeVar('LabelType', bound=Key)


class AmbiguousProvider(Exception):
    """Raised when multiple providers are found for a type."""


def _extract_typevars_from_generic_type(t: type) -> Tuple[TypeVar, ...]:
    """Returns the typevars that were used in the definition of a Generic type."""
    if not hasattr(t, '__orig_bases__'):
        return ()
    return tuple(
        chain(*(get_args(b) for b in t.__orig_bases__ if get_origin(b) == Generic))
    )


def _find_all_typevars(t: Union[type, TypeVar]) -> Set[TypeVar]:
    """Returns the set of all TypeVars in a type expression."""
    if isinstance(t, TypeVar):
        return {t}
    return set(chain(*map(_find_all_typevars, get_args(t))))


def _find_bounds_to_make_compatible_type(
    requested: Key,
    provided: Key | TypeVar,
) -> Optional[Dict[TypeVar, Key]]:
    """
    Check if a type is compatible to a provided type.
    If the types are compatible, return a mapping from typevars to concrete types
    that makes the provided type equal to the requested type.
    """
    if provided == requested:
        ret: Dict[TypeVar, Key] = {}
        return ret
    if isinstance(provided, TypeVar):
        # If the type var has no constraints, accept anything
        if not provided.__constraints__:
            return {provided: requested}
        for c in provided.__constraints__:
            if _find_bounds_to_make_compatible_type(requested, c) is not None:
                return {provided: requested}
    if get_origin(provided) is not None:
        if get_origin(provided) == get_origin(requested):
            return _find_bounds_to_make_compatible_type_tuple(
                get_args(requested), get_args(provided)
            )
    return None


def _find_bounds_to_make_compatible_type_tuple(
    requested: tuple[Key, ...],
    provided: tuple[Key | TypeVar, ...],
) -> Optional[Dict[TypeVar, Key]]:
    """
    Check if a tuple of requested types is compatible with a tuple of provided types
    and return a mapping from type vars to concrete types that makes all provided
    types equal to their corresponding requested type.
    If any of the types is not compatible, return None.
    """
    union: Dict[TypeVar, Key] = {}
    for bound in map(_find_bounds_to_make_compatible_type, requested, provided):
        # If no mapping from the type-var to a concrete type was found,
        # or if the mapping is inconsistent,
        # interrupt the search and report that no compatible types were found.
        if bound is None or any(k in union and union[k] != bound[k] for k in bound):
            return None
        union.update(bound)
    return union


def _find_all_paths(
    dependencies: Mapping[Key, Collection[Key]], start: Key, end: Key
) -> List[List[Key]]:
    """Find all paths from start to end in a DAG."""
    if start == end:
        return [[start]]
    if start not in dependencies:
        return []
    paths = []
    for node in dependencies[start]:
        for path in _find_all_paths(dependencies, node, end):
            paths.append([start] + path)
    return paths


def _find_nodes_in_paths(graph: Graph, end: Key) -> List[Key]:
    """
    Helper for Pipeline. Finds all nodes that need to be duplicated since they depend
    on a value from a param table.
    """
    start = next(iter(graph))
    dependencies = {k: tuple(p.arg_spec.keys()) for k, p in graph.items()}
    paths = _find_all_paths(dependencies, start, end)
    nodes = set()
    for path in paths:
        nodes.update(path)
    return list(nodes)


def _is_multiple_keys(
    keys: type | Iterable[type] | Item[T],
) -> bool:
    # Cannot simply use isinstance(keys, Iterable) because that is True for
    # generic aliases of iterable types, e.g.,
    #
    # class Str(sl.Scope[Param, str], str): ...
    # keys = Str[int]
    #
    # And isinstance(keys, type) does not work on its own because
    # it is False for the above type.
    return (
        not isinstance(keys, type) and not get_args(keys) and isinstance(keys, Iterable)
    )


class ReplicatorBase(Generic[IndexType]):
    def __init__(self, index_name: type, index: Iterable[IndexType], path: List[Key]):
        if len(path) == 0:
            raise UnsatisfiedRequirement(
                'Could not find path to param in param table. This is likely caused '
                'by requesting a Series that does not depend directly or transitively '
                'on any param from a table.'
            )
        self._index_name = index_name
        self.index = index
        self._path = path

    def __contains__(self, key: Key) -> bool:
        return key in self._path

    def replicate(
        self,
        key: Key,
        provider: Provider,
        get_provider: Callable[..., Tuple[Provider, Dict[TypeVar, Key]]],
    ) -> Graph:
        graph: Graph = {}
        for idx in self.index:
            subkey = self.key(idx, key)
            if isinstance(provider, _ParamSentinel):
                graph[subkey] = get_provider(subkey)[0]
            else:
                graph[subkey] = self._copy_node(key, provider, idx)
        return graph

    def _copy_node(
        self,
        key: Key,
        provider: Union[Provider, SeriesProvider[IndexType]],
        idx: IndexType,
    ) -> Provider:
        return provider.map_arg_keys(
            lambda arg: self.key(idx, arg) if arg in self else arg
        )

    def key(self, i: IndexType, value_name: Union[Type[T], Item[T]]) -> Item[T]:
        value_name = get_optional(value_name) or value_name
        label = Label(self._index_name, i)
        if isinstance(value_name, Item):
            return Item(value_name.label + (label,), value_name.tp)
        else:
            return Item((label,), value_name)


class Replicator(ReplicatorBase[IndexType]):
    r"""
    Helper for rewriting the graph to map over a given index.

    See Pipeline._build_series for context. Given a graph template, this makes a
    transformation as follows:

    S          P1[0]   P1[1]   P1[2]
    |           |       |       |
    A    ->    A[0]    A[1]    A[2]
    |           |       |       |
    B          B[0]    B[1]    B[2]

    Where S is a sentinel value, P1 are parameters from a parameter table, 0,1,2
    are indices of the param table rows, and A and B are arbitrary nodes in the graph.
    """

    def __init__(self, param_table: ParamTable, graph_template: Graph) -> None:
        index_name = param_table.row_dim
        super().__init__(
            index_name=index_name,
            index=param_table.index,
            path=_find_nodes_in_paths(graph_template, index_name),
        )


class GroupingReplicator(ReplicatorBase[LabelType], Generic[IndexType, LabelType]):
    r"""
    Helper for rewriting the graph to group by a given index.

    See Pipeline._build_series for context. Given a graph template, this makes a
    transformation as follows:

    P1[0]   P1[1]   P1[2]         P1[0]   P1[1]   P1[2]
     |       |       |             |       |       |
    A[0]    A[1]    A[2]          A[0]    A[1]    A[2]
     |       |       |             |       |       |
    B[0]    B[1]    B[2]    ->    B[0]    B[1]    B[2]
      \______|______/              \______/        |
             |                         |           |
            SB                       SB[x]       SB[y]
             |                         |           |
             C                        C[x]        C[y]

    Where SB is Series[Idx,B].  Here, the upper half of the graph originates from a
    prior transformation of a graph template using `Replicator`. The output of this
    combined with further nodes is the graph template passed to this class. x and y
    are the labels used in a grouping operation, based on the values of a ParamTable
    column P2.
    """

    def __init__(
        self, param_table: ParamTable, graph_template: Graph, label_name: type
    ) -> None:
        self._label_name = label_name
        self._group_node = self._find_grouping_node(param_table.row_dim, graph_template)
        self._groups: Dict[LabelType, List[IndexType]] = defaultdict(list)
        for idx, label in zip(param_table.index, param_table[label_name]):
            self._groups[label].append(idx)
        super().__init__(
            index_name=label_name,
            index=self._groups,
            path=_find_nodes_in_paths(graph_template, self._group_node),
        )

    def _copy_node(
        self,
        key: Key,
        provider: Union[Provider, SeriesProvider[IndexType]],
        idx: LabelType,
    ) -> Provider:
        if (not isinstance(provider, SeriesProvider)) or key != self._group_node:
            return super()._copy_node(key, provider, idx)
        labels = self._groups[idx]
        if set(labels) - set(provider.labels):
            raise ValueError(f'{labels} is not a subset of {provider.labels}')
        if tuple(provider.arg_spec.kwargs):
            raise RuntimeError(
                'A Series was provided with keyword arguments. This should not happen '
                'and is an internal error of Sciline.'
            )
        selected = {
            label: arg
            for label, arg in zip(provider.labels, provider.arg_spec.args)
            if label in labels
        }
        return SeriesProvider(selected.keys(), provider.row_dim, args=selected.values())

    def _find_grouping_node(self, index_name: Key, subgraph: Graph) -> type:
        ends: List[type] = []
        for key in subgraph:
            if get_origin(key) == Series and get_args(key)[0] == index_name:
                # Because of the succeeded get_origin we know it is a type
                ends.append(key)  # type: ignore[arg-type]
        if len(ends) == 1:
            return ends[0]
        raise ValueError(f"Could not find unique grouping node, found {ends}")


class SeriesProvider(Generic[KeyType], Provider):
    """
    Internal provider for combining results obtained based on different rows in a
    param table into a single object.
    """

    def __init__(
        self,
        labels: Iterable[KeyType],
        row_dim: Type[KeyType],
        *,
        args: Optional[Iterable[Key]] = None,
    ) -> None:
        super().__init__(
            func=self._call,
            arg_spec=ArgSpec.from_args(*(args if args is not None else labels)),
            kind='series',
        )
        self.labels = labels
        self.row_dim = row_dim

    def _call(self, *vals: ValueType) -> Series[KeyType, ValueType]:
        return Series(self.row_dim, dict(zip(self.labels, vals)))


class _ParamSentinel(Provider):
    def __init__(self, key: Key) -> None:
        super().__init__(
            func=lambda: None,
            arg_spec=ArgSpec.from_args(key),
            kind='sentinel',
            location=ProviderLocation(
                name=f'param_sentinel({type(key).__name__})', module='sciline'
            ),
        )


class Pipeline:
    """A container for providers that can be assembled into a task graph."""

    def __init__(
        self,
        providers: Optional[Iterable[Union[ToProvider, Provider]]] = None,
        *,
        params: Optional[Dict[Type[Any], Any]] = None,
    ):
        """
        Setup a Pipeline from a list providers

        Parameters
        ----------
        providers:
            List of callable providers. Each provides its return value.
            Their arguments and return value must be annotated with type hints.
        params:
            Dictionary of concrete values to provide for types.
        """
        self._providers: Dict[Key, Provider] = {}
        self._subproviders: Dict[type, Dict[Tuple[Key | TypeVar, ...], Provider]] = {}
        self._param_tables: Dict[Key, ParamTable] = {}
        self._param_name_to_table_key: Dict[Key, Key] = {}
        for provider in providers or []:
            self.insert(provider)
        for tp, param in (params or {}).items():
            self[tp] = param

    def insert(self, provider: Union[ToProvider, Provider], /) -> None:
        """
        Add a callable that provides its return value to the pipeline.

        Parameters
        ----------
        provider:
            Either a callable that provides its return value. Its arguments
            and return value must be annotated with type hints.
            Or a ``Provider`` object that has been constructed from such a callable.
        """
        if not isinstance(provider, Provider):
            provider = Provider.from_function(provider)
        self._set_provider(provider.deduce_key(), provider)

    def __setitem__(self, key: Type[T], param: T) -> None:
        """
        Provide a concrete value for a type.

        Parameters
        ----------
        key:
            Type to provide a value for.
        param:
            Concrete value to provide.
        """
        if get_origin(key) == Union:
            raise ValueError('Union (or Optional) parameters are not allowed.')
        # TODO Switch to isinstance(key, NewType) once our minimum is Python 3.10
        # Note that we cannot pass mypy in Python<3.10 since NewType is not a type.
        if hasattr(key, '__supertype__'):
            underlying = key.__supertype__  # type: ignore[attr-defined]
        else:
            underlying = key
        if (origin := get_origin(underlying)) is None:
            # In Python 3.8, get_origin does not work with numpy.typing.NDArray,
            # but it defines __origin__
            if (np_origin := getattr(underlying, '__origin__', None)) is not None:
                expected = np_origin
            else:
                expected = underlying
        elif issubclass(origin, (Scope, ScopeTwoParams)):
            scope = origin.__orig_bases__[0]
            while (orig := get_origin(scope)) is not None and orig not in (
                Scope,
                ScopeTwoParams,
            ):
                scope = orig.__orig_bases__[0]
            expected = get_args(scope)[-1]
        else:
            expected = origin

        if not isinstance(param, expected):
            raise TypeError(
                f'Key {key} incompatible to value {param} of type {type(param)}'
            )
        self._set_provider(key, Provider.parameter(param))

    def set_param_table(self, params: ParamTable) -> None:
        """
        Set a parameter table for a row dimension.

        Values in the parameter table provide concrete values for a type given by the
        respective column header.

        A pipeline can have multiple parameter tables, but only one per row dimension.
        Column names must be unique across all parameter tables.

        Parameters
        ----------
        params:
            Parameter table to set.
        """
        for param_name in params:
            if (existing := self._param_name_to_table_key.get(param_name)) is not None:
                if (
                    existing == params.row_dim
                    and param_name in self._param_tables[existing]
                ):
                    # Column will be removed by del_param_table below, clash is ok
                    continue
                raise ValueError(f'Parameter {param_name} already set')
        if params.row_dim in self._param_tables:
            self.del_param_table(params.row_dim)
        self._param_tables[params.row_dim] = params
        for param_name in params:
            self._param_name_to_table_key[param_name] = params.row_dim
        for param_name, values in params.items():
            for index, label in zip(params.index, values):
                self._set_provider(
                    Item((Label(tp=params.row_dim, index=index),), param_name),
                    Provider.table_cell(label),
                )
        for index, label in zip(params.index, params.index):
            self._set_provider(
                Item((Label(tp=params.row_dim, index=index),), params.row_dim),
                Provider.table_cell(label),
            )

    def del_param_table(self, row_dim: type) -> None:
        """
        Remove a parameter table.

        Parameters
        ----------
        row_dim:
            Row dimension of the parameter table to remove.
        """
        # 1. Remove providers pointing to table cells
        params = self._param_tables[row_dim]
        for index in params.index:
            label = (Label(tp=row_dim, index=index),)
            for param_name in params:
                del self._providers[Item(label, param_name)]
            del self._providers[Item(label, row_dim)]
        # 2. Remove column to table mapping
        for param_name in list(self._param_name_to_table_key):
            if self._param_name_to_table_key[param_name] == row_dim:
                del self._param_name_to_table_key[param_name]
        # 3. Remove table
        del self._param_tables[row_dim]

    def set_param_series(self, row_dim: type, index: Collection[Any]) -> None:
        """
        Set a series of parameters.

        This is a convenience method for creating and setting a parameter table with
        no columns and an index given by `index`.

        Parameters
        ----------
        row_dim:
            Row dimension of the parameter table to set.
        index:
            Index of the parameter table to set.
        """
        self.set_param_table(ParamTable(row_dim, columns={}, index=index))

    def _set_provider(
        self,
        key: Key,
        provider: Provider,
    ) -> None:
        # isinstance does not work here and types.NoneType available only in 3.10+
        if key == type(None):  # noqa: E721
            raise ValueError(f'Provider {provider} returning `None` is not allowed')
        if get_origin(key) == Union:
            raise ValueError(
                f'Provider {provider} returning a Union (or Optional) is not allowed.'
            )
        if get_origin(key) == Series:
            raise ValueError(
                f'Provider {provider} returning a sciline.Series is not allowed. '
                'Series is a special container reserved for use in conjunction with '
                'sciline.ParamTable and must not be provided directly.'
            )
        if (origin := get_origin(key)) is not None:
            subproviders = self._subproviders.setdefault(origin, {})
            args = get_args(key)
            subproviders[args] = provider
        else:
            self._providers[key] = provider

    def _get_provider(
        self, tp: Union[Type[T], Item[T]], handler: Optional[ErrorHandler] = None
    ) -> Tuple[Provider, Dict[TypeVar, Key]]:
        handler = handler or HandleAsBuildTimeException()
        explanation: List[str] = []
        if (provider := self._providers.get(tp)) is not None:
            return provider, {}
        elif (origin := get_origin(tp)) is not None and (
            subproviders := self._subproviders.get(origin)
        ) is not None:
            requested = get_args(tp)
            matches = [
                (subprovider, bound)
                for args, subprovider in subproviders.items()
                if (
                    bound := _find_bounds_to_make_compatible_type_tuple(requested, args)
                )
                is not None
            ]
            typevar_counts = [len(bound) for _, bound in matches]
            min_typevar_count = min(typevar_counts, default=0)
            matches = [
                m
                for count, m in zip(typevar_counts, matches)
                if count == min_typevar_count
            ]

            if len(matches) == 1:
                provider, bound = matches[0]
                return provider, bound
            elif len(matches) > 1:
                matching_providers = [provider.location.name for provider, _ in matches]
                raise AmbiguousProvider(
                    f"Multiple providers found for type {tp}."
                    f" Matching providers are: {matching_providers}."
                )
            else:
                typevars_in_expression = _extract_typevars_from_generic_type(origin)
                if typevars_in_expression:
                    explanation = [
                        ''.join(
                            map(
                                str,
                                (
                                    'Note that ',
                                    keyname(origin[typevars_in_expression]),
                                    ' has constraints ',
                                    (
                                        {
                                            keyname(tv): tuple(
                                                map(keyname, tv.__constraints__)
                                            )
                                            for tv in typevars_in_expression
                                        }
                                    ),
                                ),
                            )
                        )
                    ]
        return handler.handle_unsatisfied_requirement(tp, *explanation), {}

    def _get_unique_provider(
        self, tp: Union[Type[T], Item[T]], handler: ErrorHandler
    ) -> Tuple[Provider, Dict[TypeVar, Key]]:
        """Get a unique provider for a potential Union type."""
        if (union_args := get_union(tp)) is None:
            return self._get_provider(tp, handler=handler)
        matching_types = []
        for option in union_args:
            try:
                provider, bound = self._get_provider(option, handler=None)
            except UnsatisfiedRequirement:
                continue
            else:
                matching_types.append(option)
        if len(matching_types) == 0:
            return handler.handle_unsatisfied_requirement(tp), {}
        if len(matching_types) > 1:
            raise AmbiguousProvider(
                f"Multiple providers found for Union type {tp}."
                f" Matching types are: {matching_types}."
            )
        return provider, bound

    def build(
        self,
        tp: Union[Type[T], Item[T]],
        /,
        *,
        handler: ErrorHandler,
        search_param_tables: bool = False,
    ) -> Graph:
        """
        Return a dict of providers required for building the requested type `tp`.

        This is mainly for internal and low-level use. Prefer using :py:meth:`get`.

        The values are tuples containing the provider and the dict of arguments for
        the provider. The values in the latter dict reference other keys in the returned
        graph.

        Parameters
        ----------
        tp:
            Type to build the graph for.
        search_param_tables:
            Whether to search parameter tables for concrete keys.
        """
        graph: Graph = {}
        stack: List[Union[Type[T], Item[T]]] = [tp]
        while stack:
            tp = stack.pop()
            # First look in column labels of param tables
            if search_param_tables and tp in self._param_name_to_table_key:
                graph[tp] = _ParamSentinel(self._param_name_to_table_key[tp])
                continue
            # Then also indices of param tables. This comes second because we need to
            # prefer column labels over indices for multi-level grouping.
            if search_param_tables and tp in self._param_tables:
                graph[tp] = _ParamSentinel(tp)
                continue
            if get_origin(tp) == Series:
                sub = self._build_series(tp, handler=handler)  # type: ignore[arg-type]
                graph.update(sub)
                continue
            if (optional_arg := get_optional(tp)) is not None:
                try:
                    optional_subgraph = self.build(
                        optional_arg,
                        search_param_tables=search_param_tables,
                        handler=HandleAsBuildTimeException(),
                    )
                except UnsatisfiedRequirement:
                    graph[tp] = Provider.provide_none()
                else:
                    graph[tp] = optional_subgraph.pop(optional_arg)
                    graph.update(optional_subgraph)
                continue
            provider, bound = self._get_unique_provider(tp, handler=handler)
            provider = provider.bind_type_vars(bound)
            graph[tp] = provider
            stack.extend(provider.arg_spec.keys() - graph.keys())
        return graph

    def _build_series(
        self, tp: Type[Series[KeyType, ValueType]], handler: ErrorHandler
    ) -> Graph:
        """
        Build (sub)graph for a Series type implementing ParamTable-based functionality.

        We illustrate this with an example. Given a ParamTable with row_dim 'Idx':

        Idx | P1 | P2
        0   | a  | x
        1   | b  | x
        2   | c  | y

        and providers for A depending on P1 and B depending on A.  Calling
        build(Series[Idx,B]) will call _build_series(Series[Idx,B]). This results in
        the following procedure here:

        1. Call build(P1), resulting, e.g., in a graph S->A->B, where S is a sentinel.
           The sentinel is used because build() cannot find a unique P1, since it is
           not a single value but a column in a table.
        2. Instantiation of `Replicator`, which will be used to replicate the
           relevant parts of the graph (see illustration there).
        3. Insert a special `SeriesProvider` node, which will gather the duplicates of
           the 'B' node and provides the requested Series[Idx,B].
        4. Replicate the graph. Nodes that do not directly or indirectly depend on P1
           are not replicated.

        Conceptually, the final result will be {
            0: B(A(a)),
            1: B(A(b)),
            2: B(A(c))
        }.

        In more complex cases, we may be dealing with multiple levels of Series,
        which is used for grouping operations. Consider the above example, but with
        and additional provider for C depending on Series[Idx,B]. Calling
        build(Series[P2,C]) will call _build_series(Series[P2,C]). This results in
        the following procedure here:

        a. Call build(C), which results in the procedure above, i.e., a nested call
           to _build_series(Series[Idx,B]) and the resulting graph as explained above.
        b. Instantiation of `GroupingReplicator`, which will be used to replicate the
           relevant parts of the graph (see illustration there).
        c. Insert a special `SeriesProvider` node, which will gather the duplicates of
           the 'C' node and providers the requested Series[P2,C].
        c. Replicate the graph. Nodes that do not directly or indirectly depend on
           the special `SeriesProvider` node (from step 3.) are not replicated.

        Conceptually, the final result will be {
            x: C({
                0: B(A(a)),
                1: B(A(b))
            }),
            y: C({
                2: B(A(c))
            })
        }.
        """
        index_name: Type[KeyType]
        value_type: Type[ValueType]
        index_name, value_type = get_args(tp)

        subgraph = self.build(value_type, search_param_tables=True, handler=handler)

        replicator: ReplicatorBase[KeyType]
        if (
            # For multi-level grouping a type is an index as well as a column label.
            # In this case we do not want to replicate the graph, but group it (elif).
            index_name not in self._param_name_to_table_key
            and (params := self._param_tables.get(index_name)) is not None
        ):
            replicator = Replicator(param_table=params, graph_template=subgraph)
        elif (table_key := self._param_name_to_table_key.get(index_name)) is not None:
            replicator = GroupingReplicator(
                param_table=self._param_tables[table_key],
                graph_template=subgraph,
                label_name=index_name,
            )
        else:
            raise KeyError(f'No parameter table found for label {index_name}')

        graph: Graph = {
            tp: SeriesProvider(
                list(replicator.index),
                index_name,
                args=(replicator.key(idx, value_type) for idx in replicator.index),
            )
        }

        for key, provider in subgraph.items():
            if key in replicator:
                graph.update(replicator.replicate(key, provider, self._get_provider))
            else:
                graph[key] = provider
        return graph

    @overload
    def compute(self, tp: Type[T], **kwargs: Any) -> T:
        ...

    @overload
    def compute(self, tp: Iterable[Type[T]], **kwargs: Any) -> Dict[Type[T], T]:
        ...

    @overload
    def compute(self, tp: Item[T], **kwargs: Any) -> T:
        ...

    def compute(self, tp: type | Iterable[type] | Item[T], **kwargs: Any) -> Any:
        """
        Compute result for the given keys.

        Equivalent to ``self.get(tp).compute()``.

        Parameters
        ----------
        tp:
            Type to compute the result for.
            Can be a single type or an iterable of types.
        kwargs:
            Keyword arguments passed to the ``.get()`` method.
        """
        return self.get(tp, **kwargs).compute()

    def visualize(
        self, tp: type | Iterable[type], **kwargs: Any
    ) -> graphviz.Digraph:  # type: ignore[name-defined] # noqa: F821
        """
        Return a graphviz Digraph object representing the graph for the given keys.

        Equivalent to ``self.get(tp).visualize()``.

        Parameters
        ----------
        tp:
            Type to visualize the graph for.
            Can be a single type or an iterable of types.
        kwargs:
            Keyword arguments passed to :py:class:`graphviz.Digraph`.
        """
        return self.get(tp, handler=HandleAsComputeTimeException()).visualize(**kwargs)

    def get(
        self,
        keys: type | Iterable[type] | Item[T],
        *,
        scheduler: Optional[Scheduler] = None,
        handler: Optional[ErrorHandler] = None,
    ) -> TaskGraph:
        """
        Return a TaskGraph for the given keys.

        Parameters
        ----------
        keys:
            Type to compute the result for.
            Can be a single type or an iterable of types.
        scheduler:
            Optional scheduler to use for computing the result. If not given, a
            :py:class:`NaiveScheduler` is used if `dask` is not installed,
            otherwise dask's threaded scheduler is used.
        handler:
            Handler for unsatisfied requirements. If not provided,
            :py:class:`HandleAsBuildTimeException` is used, which raises an exception.
            During development and debugging it can be helpful to use a handler that
            raises an exception only when the graph is computed. This can be achieved
            by passing :py:class:`HandleAsComputeTimeException` as the handler.
        """
        handler = handler or HandleAsBuildTimeException()
        if _is_multiple_keys(keys):
            keys = tuple(keys)  # type: ignore[arg-type]
            graph: Graph = {}
            for t in keys:
                graph.update(self.build(t, handler=handler))
        else:
            graph = self.build(keys, handler=handler)  # type: ignore[arg-type]
        return TaskGraph(
            graph=graph, targets=keys, scheduler=scheduler  # type: ignore[arg-type]
        )

    @overload
    def bind_and_call(self, fns: Callable[..., T], /) -> T:
        ...

    @overload
    def bind_and_call(self, fns: Iterable[Callable[..., Any]], /) -> Tuple[Any, ...]:
        ...

    def bind_and_call(
        self, fns: Union[Callable[..., Any], Iterable[Callable[..., Any]]], /
    ) -> Any:
        """
        Call the given functions with arguments provided by the pipeline.

        Parameters
        ----------
        fns:
            Functions to call.
            The pipeline will provide all arguments based on the function's type hints.

            If this is a single callable, it is called directly.
            Otherwise, ``bind_and_call`` will iterate over it and call all functions.
            If will in either case call :meth:`Pipeline.compute` only once.

        Returns
        -------
        :
            The return values of the functions in the same order as the functions.
            If only one function is passed, its return value
            is *not* wrapped in a tuple.
        """
        return_tuple = True
        if callable(fns):
            fns = (fns,)
            return_tuple = False

        arg_types_per_function = {
            fn: {
                name: ty for name, ty in get_type_hints(fn).items() if name != 'return'
            }
            for fn in fns
        }
        all_arg_types = tuple(
            set(chain(*(a.values() for a in arg_types_per_function.values())))
        )
        values_per_type = self.compute(all_arg_types)
        results = tuple(
            fn(**{name: values_per_type[ty] for name, ty in arg_types.items()})
            for fn, arg_types in arg_types_per_function.items()
        )
        if not return_tuple:
            return results[0]
        return results

    def copy(self) -> Pipeline:
        """
        Make a copy of the pipeline.
        """
        out = Pipeline()
        out._providers = self._providers.copy()
        out._subproviders = {k: v.copy() for k, v in self._subproviders.items()}
        out._param_tables = self._param_tables.copy()
        out._param_name_to_table_key = self._param_name_to_table_key.copy()
        return out

    def __copy__(self) -> Pipeline:
        return self.copy()

    def _repr_html_(self) -> str:
        providers_without_parameters = (
            (origin, tuple(), value) for origin, value in self._providers.items()
        )  # type: ignore[var-annotated]
        providers_with_parameters = (
            (origin, args, value)
            for origin in self._subproviders
            for args, value in self._subproviders[origin].items()
        )
        return pipeline_html_repr(
            chain(providers_without_parameters, providers_with_parameters)
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# Tests on the hidden provider module

from typing import List, Tuple, TypeVar

from sciline._provider import ArgSpec


def test_arg_spec() -> None:
    arg_spec = ArgSpec(args={'a': int}, kwargs={'b': str})
    assert list(arg_spec.args) == [int]
    assert dict(arg_spec.kwargs) == {'b': str}


def combine_numbers(a: int, *, b: float) -> str:
    return f"{a} and {b}"


T = TypeVar('T', int, float)


def complicated_append(a: T, *, b: List[T]) -> Tuple[T, ...]:
    b.append(a)

    return tuple(b)


def test_arg_spec_from_function_simple() -> None:
    arg_spec = ArgSpec.from_function(combine_numbers)
    assert list(arg_spec.args) == [int]
    assert dict(arg_spec.kwargs) == {'b': float}


def test_arg_spec_from_function_typevar() -> None:
    arg_spec = ArgSpec.from_function(complicated_append)

    assert list(arg_spec.args) == [T]
    assert dict(arg_spec.kwargs) == {'b': List[T]}  # type: ignore[valid-type]
    specific_arg_spec = arg_spec.bind_type_vars(
        bound={T: int}  # type: ignore[dict-item]
    )
    assert list(specific_arg_spec.args) == [int]
    assert dict(specific_arg_spec.kwargs) == {'b': list[int]}


def test_arg_spec_decorated_function_with_wraps() -> None:
    from typing import Callable, Union

    def decorator(func: Callable[..., str]) -> Callable[..., str]:
        from functools import wraps

        @wraps(func)
        def wrapper(*args: Union[int, float], **kwargs: Union[int, float]) -> str:
            return "Wrapped: " + func(*args, **kwargs)

        return wrapper

    @decorator
    def decorated(a: int, *, b: float) -> str:
        return f"{a} and {b}"

    arg_spec = ArgSpec.from_function(decorated)
    assert list(arg_spec.args) == [int]
    assert dict(arg_spec.kwargs) == {'b': float}


def test_arg_spec_decorated_function_without_wraps() -> None:
    from typing import Callable, Union

    def decorator(func: Callable[..., str]) -> Callable[..., str]:
        def wrapper(*args: Union[int, float], **kwargs: Union[int, float]) -> str:
            return "Wrapped: " + func(*args, **kwargs)

        return wrapper

    @decorator
    def decorated(a: int, *, b: float) -> str:
        return f"{a} and {b}"

    arg_spec = ArgSpec.from_function(decorated)
    assert list(arg_spec.args) == []
    assert dict(arg_spec.kwargs) == {}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import inspect
from typing import Any, Callable, Dict, List, Optional, Protocol, Tuple

from sciline.typing import Graph, Key


class CycleError(Exception):
    pass


class Scheduler(Protocol):
    """
    Scheduler interface compatible with :py:class:`sciline.Pipeline`.
    """

    def get(self, graph: Graph, keys: List[Key]) -> Tuple[Any, ...]:
        """
        Compute the result for given keys from the graph.

        Must raise :py:class:`sciline.scheduler.CycleError` if the graph contains
        a cycle.
        """
        ...


class NaiveScheduler:
    """
    A naive scheduler that computes intermediate results and results in order.

    May consume excessive memory since intermediate results are not freed eagerly,
    but kept until returning the final result. Prefer installing `dask` and using
    :py:class:`DaskScheduler` instead.
    """

    def get(self, graph: Graph, keys: List[Key]) -> Tuple[Any, ...]:
        import graphlib

        dependencies = {
            tp: tuple(provider.arg_spec.keys()) for tp, provider in graph.items()
        }
        ts = graphlib.TopologicalSorter(dependencies)
        try:
            # Create list from generator to force early exception if there is a cycle
            tasks = list(ts.static_order())
        except graphlib.CycleError as e:
            raise CycleError from e
        results: Dict[Key, Any] = {}
        for t in tasks:
            results[t] = graph[t].call(results)
        return tuple(results[key] for key in keys)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'


class DaskScheduler:
    """Wrapper for a Dask scheduler.

    Note that this currently only works if all providers support posargs.
    """

    def __init__(self, scheduler: Optional[Callable[..., Any]] = None) -> None:
        """Wrap a dask scheduler or the default `dask.threaded.get`.

        Parameters
        ----------
        scheduler:
            A Dask scheduler, such as `dask.get`, `dask.threaded.get`,
            `dask.multiprocessing.get, or `dask.distributed.Client.get`.
        """
        if scheduler is None:
            import dask

            self._dask_get = dask.threaded.get
        else:
            self._dask_get = scheduler

    def get(self, graph: Graph, keys: List[Key]) -> Any:
        from dask.utils import apply

        # Use `apply` to allow passing keyword arguments.
        # Contrary to the Dask docs, we need to pass positional args as a list
        # and keyword args with a nested tuple+list structure.
        # Otherwise, Dask would treat them as literal values instead of
        # references to other nodes.
        dsk = {
            tp: (
                apply,
                provider.func,
                list(provider.arg_spec.args),
                (dict, [[key, val] for key, val in provider.arg_spec.kwargs]),
            )
            for tp, provider in graph.items()
        }
        try:
            return self._dask_get(dsk, keys)
        except RuntimeError as e:
            if str(e).startswith("Cycle detected"):
                raise CycleError from e
            raise

    def __repr__(self) -> str:
        module = getattr(inspect.getmodule(self._dask_get), '__name__', '')
        name = self._dask_get.__name__
        return f'{self.__class__.__name__}({module}.{name})'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import functools
from dataclasses import dataclass
from typing import Any, Callable, Generic, List, NewType, TypeVar

import numpy as np
import numpy.typing as npt
import pytest

import sciline as sl


def int_to_float(x: int) -> float:
    return 0.5 * x


def make_int() -> int:
    return 3


def int_float_to_str(x: int, y: float) -> str:
    return f"{x};{y}"


def test_pipeline_with_callables_can_compute_single_results() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    assert pipeline.compute(float) == 1.5
    assert pipeline.compute(int) == 3


def test_pipeline_does_not_autobind_types_that_can_be_default_constructed() -> None:
    # `int` can be constructed without arguments (and returns 0). Make sure that
    # the pipeline does not automatically bind `int` to `0`.
    pipeline = sl.Pipeline([int_to_float])
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.compute(float)


def test_intermediate_used_multiple_times_is_computed_only_once() -> None:
    ncall = 0

    def provide_int() -> int:
        nonlocal ncall
        ncall += 1
        return 3

    pipeline = sl.Pipeline([int_to_float, provide_int, int_float_to_str])
    assert pipeline.compute(str) == "3;1.5"
    assert ncall == 1


def test_multiple_keys_can_be_computed_without_repeated_calls() -> None:
    ncall = 0

    def provide_int() -> int:
        nonlocal ncall
        ncall += 1
        return 3

    pipeline = sl.Pipeline([int_to_float, provide_int, int_float_to_str])
    assert pipeline.compute((float, str)) == {float: 1.5, str: "3;1.5"}
    assert ncall == 1


def test_multiple_keys_not_in_same_path_use_same_intermediate() -> None:
    ncall = 0

    def provide_int() -> int:
        nonlocal ncall
        ncall += 1
        return 3

    def func1(x: int) -> float:
        return 0.5 * x

    def func2(x: int) -> str:
        return f"{x}"

    pipeline = sl.Pipeline([provide_int, func1, func2])
    assert pipeline.compute((float, str)) == {float: 1.5, str: "3"}
    assert ncall == 1


def test_Scope_subclass_can_be_set_as_param() -> None:
    Param = TypeVar('Param')

    class Str(sl.Scope[Param, str], str):
        ...

    pipeline = sl.Pipeline(params={Str[int]: Str[int]('1')})
    pipeline[Str[float]] = Str[float]('2.0')
    assert pipeline.compute(Str[int]) == Str[int]('1')
    assert pipeline.compute(Str[float]) == Str[float]('2.0')


def test_Scope_subclass_can_be_set_as_param_with_unbound_typevar() -> None:
    Param = TypeVar('Param')

    class Str(sl.Scope[Param, str], str):
        ...

    pipeline = sl.Pipeline()
    pipeline[Str[Param]] = Str[Param]('1')  # type: ignore[valid-type]
    assert pipeline.compute(Str[int]) == Str[int]('1')
    assert pipeline.compute(Str[float]) == Str[float]('1')


def test_ScopeTwoParam_subclass_can_be_set_as_param() -> None:
    Param1 = TypeVar('Param1')
    Param2 = TypeVar('Param2')

    class Str(sl.ScopeTwoParams[Param1, Param2, str], str):
        ...

    pipeline = sl.Pipeline(params={Str[int, float]: Str[int, float]('1')})
    pipeline[Str[float, int]] = Str[float, int]('2.0')
    assert pipeline.compute(Str[int, float]) == Str[int, float]('1')
    assert pipeline.compute(Str[float, int]) == Str[float, int]('2.0')


def test_ScopeTwoParam_subclass_can_be_set_as_param_with_unbound_typevar() -> None:
    Param1 = TypeVar('Param1')
    Param2 = TypeVar('Param2')

    class Str(sl.ScopeTwoParams[Param1, Param2, str], str):
        ...

    pipeline = sl.Pipeline()
    pipeline[Str[Param1, Param2]] = Str[Param1, Param2]('1')  # type: ignore[valid-type]
    assert pipeline.compute(Str[int, float]) == Str[int, float]('1')
    assert pipeline.compute(Str[float, int]) == Str[float, int]('1')


def test_generic_providers_produce_use_dependencies_based_on_bound_typevar() -> None:
    Param = TypeVar('Param')

    class Str(sl.Scope[Param, str], str):
        ...

    def parametrized(x: Param) -> Str[Param]:
        return Str(f'{x}')

    def make_float() -> float:
        return 1.5

    def combine(x: Str[int], y: Str[float]) -> str:
        return f"{x};{y}"

    pipeline = sl.Pipeline([make_int, make_float, combine, parametrized])
    assert pipeline.compute(Str[int]) == Str[int]('3')
    assert pipeline.compute(Str[float]) == Str[float]('1.5')
    assert pipeline.compute(str) == '3;1.5'


def test_can_compute_result_depending_on_two_instances_of_generic_provider() -> None:
    ncall = 0

    def provide_int() -> int:
        nonlocal ncall
        ncall += 1
        return 3

    Param = TypeVar('Param')

    class Float(sl.Scope[Param, float], float):
        ...

    class Str(sl.Scope[Param, str], str):
        ...

    def int_float_to_str(x: int, y: Float[Param]) -> Str[Param]:
        return Str(f"{x};{y}")

    Run1 = NewType('Run1', int)
    Run2 = NewType('Run2', int)
    Result = NewType('Result', str)

    def float1() -> Float[Run1]:
        return Float[Run1](1.5)

    def float2() -> Float[Run2]:
        return Float[Run2](2.5)

    def use_strings(s1: Str[Run1], s2: Str[Run2]) -> Result:
        return Result(f"{s1};{s2}")

    pipeline = sl.Pipeline(
        [provide_int, float1, float2, use_strings, int_float_to_str],
    )
    assert pipeline.compute(Result) == "3;1.5;3;2.5"
    assert ncall == 1


def test_subclasses_of_generic_provider_defined_with_Scope_work() -> None:
    Param = TypeVar('Param')

    class StrT(sl.Scope[Param, str], str):
        ...

    class Str1(StrT[Param]):
        ...

    class Str2(StrT[Param]):
        ...

    class Str3(StrT[Param]):
        ...

    class Str4(Str3[Param]):
        ...

    def make_str1() -> Str1[Param]:
        return Str1('1')

    def make_str2() -> Str2[Param]:
        return Str2('2')

    # Note that mypy cannot detect if when setting params, the type of the
    # parameter does not match the key. Same problem as with NewType.
    pipeline = sl.Pipeline(
        [make_str1, make_str2],
        params={
            Str3[int]: Str3[int]('int3'),
            Str3[float]: Str3[float]('float3'),
            Str4[int]: Str2[int]('int4'),
        },
    )
    assert pipeline.compute(Str1[float]) == Str1[float]('1')
    assert pipeline.compute(Str2[float]) == Str2[float]('2')
    assert pipeline.compute(Str3[int]) == Str3[int]('int3')
    assert pipeline.compute(Str3[float]) == Str3[float]('float3')
    assert pipeline.compute(Str4[int]) == Str4[int]('int4')


def test_subclasses_of_generic_array_provider_defined_with_Scope_work() -> None:
    Param = TypeVar('Param')

    class ArrayT(sl.Scope[Param, npt.NDArray[np.int64]], npt.NDArray[np.int64]):
        ...

    class Array1(ArrayT[Param]):
        ...

    class Array2(ArrayT[Param]):
        ...

    def make_array1() -> Array1[Param]:
        return Array1(np.array([1, 2, 3]))

    def make_array2() -> Array2[Param]:
        return Array2(np.array([4, 5, 6]))

    pipeline = sl.Pipeline([make_array1, make_array2])
    # Note that the param is not the dtype
    assert np.all(pipeline.compute(Array1[str]) == np.array([1, 2, 3]))
    assert np.all(pipeline.compute(Array2[str]) == np.array([4, 5, 6]))


def test_inserting_provider_returning_None_raises() -> None:
    def provide_none() -> None:
        return None

    with pytest.raises(ValueError):
        sl.Pipeline([provide_none])
    pipeline = sl.Pipeline([])
    with pytest.raises(ValueError):
        pipeline.insert(provide_none)


def test_inserting_provider_with_no_return_type_raises() -> None:
    def provide_none():  # type: ignore[no-untyped-def]
        return None

    with pytest.raises(ValueError):
        sl.Pipeline([provide_none])
    pipeline = sl.Pipeline([])
    with pytest.raises(ValueError):
        pipeline.insert(provide_none)


def test_TypeVar_requirement_of_provider_can_be_bound() -> None:
    T = TypeVar('T')

    def provider_int() -> int:
        return 3

    def provider(x: T) -> List[T]:
        return [x, x]

    pipeline = sl.Pipeline([provider_int, provider])
    assert pipeline.compute(List[int]) == [3, 3]


def test_TypeVar_that_cannot_be_bound_raises_UnboundTypeVar() -> None:
    T = TypeVar('T')

    def provider(_: T) -> int:
        return 1

    pipeline = sl.Pipeline([provider])
    with pytest.raises(sl.UnboundTypeVar):
        pipeline.compute(int)


def test_unsatisfiable_TypeVar_requirement_of_provider_raises() -> None:
    T = TypeVar('T')

    def provider_int() -> int:
        return 3

    def provider(x: T) -> List[T]:
        return [x, x]

    pipeline = sl.Pipeline([provider_int, provider])
    with pytest.raises(sl.UnsatisfiedRequirement):
        pipeline.compute(List[float])


def test_TypeVar_params_are_not_associated_unless_they_match() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    class A(Generic[T1]):
        ...

    class B(Generic[T2]):
        ...

    def source() -> A[int]:
        return A[int]()

    def not_matching(x: A[T1]) -> B[T2]:
        return B[T2]()

    def matching(x: A[T1]) -> B[T1]:
        return B[T1]()

    pipeline = sl.Pipeline([source, not_matching])
    with pytest.raises(sl.UnboundTypeVar):
        pipeline.compute(B[int])

    pipeline = sl.Pipeline([source, matching])
    pipeline.compute(B[int])


def test_multi_Generic_with_fully_bound_arguments() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1, T2]):
        first: T1
        second: T2

    def source() -> A[int, float]:
        return A[int, float](1, 2.0)

    pipeline = sl.Pipeline([source])
    assert pipeline.compute(A[int, float]) == A[int, float](1, 2.0)


def test_multi_Generic_with_partially_bound_arguments() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1, T2]):
        first: T1
        second: T2

    def source() -> float:
        return 2.0

    def partially_bound(x: T1) -> A[int, T1]:
        return A[int, T1](1, x)

    pipeline = sl.Pipeline([source, partially_bound])
    assert pipeline.compute(A[int, float]) == A[int, float](1, 2.0)


def test_multi_Generic_with_multiple_unbound() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1, T2]):
        first: T1
        second: T2

    def int_source() -> int:
        return 1

    def float_source() -> float:
        return 2.0

    def unbound(x: T1, y: T2) -> A[T1, T2]:
        return A[T1, T2](x, y)

    pipeline = sl.Pipeline([int_source, float_source, unbound])
    assert pipeline.compute(A[int, float]) == A[int, float](1, 2.0)
    assert pipeline.compute(A[float, int]) == A[float, int](2.0, 1)


def test_distinct_fully_bound_instances_yield_distinct_results() -> None:
    T1 = TypeVar('T1')

    @dataclass
    class A(Generic[T1]):
        value: T1

    def int_source() -> A[int]:
        return A[int](1)

    def float_source() -> A[float]:
        return A[float](2.0)

    pipeline = sl.Pipeline([int_source, float_source])
    assert pipeline.compute(A[int]) == A[int](1)
    assert pipeline.compute(A[float]) == A[float](2.0)


def test_distinct_partially_bound_instances_yield_distinct_results() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1, T2]):
        first: T1
        second: T2

    def str_source() -> str:
        return 'a'

    def int_source(x: T1) -> A[int, T1]:
        return A[int, T1](1, x)

    def float_source(x: T1) -> A[float, T1]:
        return A[float, T1](2.0, x)

    pipeline = sl.Pipeline([str_source, int_source, float_source])
    assert pipeline.compute(A[int, str]) == A[int, str](1, 'a')
    assert pipeline.compute(A[float, str]) == A[float, str](2.0, 'a')


def test_multiple_matching_partial_providers_raises() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1, T2]):
        first: T1
        second: T2

    def int_source() -> int:
        return 1

    def float_source() -> float:
        return 2.0

    def provider1(x: T1) -> A[int, T1]:
        return A[int, T1](1, x)

    def provider2(x: T2) -> A[T2, float]:
        return A[T2, float](x, 2.0)

    pipeline = sl.Pipeline([int_source, float_source, provider1, provider2])
    assert pipeline.compute(A[int, int]) == A[int, int](1, 1)
    assert pipeline.compute(A[float, float]) == A[float, float](2.0, 2.0)
    with pytest.raises(sl.AmbiguousProvider):
        pipeline.compute(A[int, float])


def test_TypeVar_params_track_to_multiple_sources() -> None:
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class A(Generic[T1]):
        value: T1

    @dataclass
    class B(Generic[T1]):
        value: T1

    @dataclass
    class C(Generic[T1, T2]):
        first: T1
        second: T2

    def provide_int() -> int:
        return 1

    def provide_float() -> float:
        return 2.0

    def provide_A(x: T1) -> A[T1]:
        return A[T1](x)

    # Note that it currently does not matter which TypeVar instance we use here:
    # Container tracks uses of TypeVar within a single provider, but does not carry
    # the information beyond the scope of a single call.
    def provide_B(x: T1) -> B[T1]:
        return B[T1](x)

    def provide_C(x: A[T1], y: B[T2]) -> C[T1, T2]:
        return C[T1, T2](x.value, y.value)

    pipeline = sl.Pipeline(
        [provide_int, provide_float, provide_A, provide_B, provide_C]
    )
    assert pipeline.compute(C[int, float]) == C[int, float](1, 2.0)


def test_instance_provider() -> None:
    Result = NewType('Result', float)

    def f(x: int, y: float) -> Result:
        return Result(x / y)

    pl = sl.Pipeline([f])
    pl[int] = 3
    pl[float] = 2.0
    assert pl.compute(int) == 3
    assert pl.compute(float) == 2.0
    assert pl.compute(Result) == 1.5


def test_provider_NewType_instance() -> None:
    A = NewType('A', int)
    pl = sl.Pipeline([])
    pl[A] = A(3)
    assert pl.compute(A) == 3


def test_setitem_generic_sets_up_working_subproviders() -> None:
    T = TypeVar('T')

    @dataclass
    class A(Generic[T]):
        value: T

    pl = sl.Pipeline()
    pl[A[int]] = A[int](3)
    pl[A[float]] = A[float](2.0)
    assert pl.compute(A[int]) == A[int](3)
    assert pl.compute(A[float]) == A[float](2.0)
    with pytest.raises(sl.UnsatisfiedRequirement):
        pl.compute(A[str])


def test_setitem_generic_works_without_params() -> None:
    T = TypeVar('T')

    @dataclass
    class A(Generic[T]):
        value: T

    pl = sl.Pipeline()
    pl[A] = A(3)
    assert pl.compute(A) == A(3)


def test_setitem_raises_TypeError_if_instance_does_not_match_key() -> None:
    A = NewType('A', int)
    T = TypeVar('T')

    @dataclass
    class B(Generic[T]):
        value: T

    pl = sl.Pipeline()
    with pytest.raises(TypeError):
        pl[int] = 1.0
    with pytest.raises(TypeError):
        pl[A] = 1.0
    with pytest.raises(TypeError):
        pl[B[int]] = 1.0


def test_setitem_can_replace_param_with_param() -> None:
    pl = sl.Pipeline()
    pl[int] = 1
    pl[int] = 2
    assert pl.compute(int) == 2


def test_insert_can_replace_param_with_provider() -> None:
    def func() -> int:
        return 2

    pl = sl.Pipeline()
    pl[int] = 1
    pl.insert(func)
    assert pl.compute(int) == 2


def test_setitem_can_replace_provider_with_param() -> None:
    def func() -> int:
        return 2

    pl = sl.Pipeline()
    pl.insert(func)
    pl[int] = 1
    assert pl.compute(int) == 1


def test_insert_can_replace_provider_with_provider() -> None:
    def func1() -> int:
        return 1

    def func2() -> int:
        return 2

    pl = sl.Pipeline()
    pl.insert(func1)
    pl.insert(func2)
    assert pl.compute(int) == 2


def test_insert_can_replace_generic_provider_with_generic_provider() -> None:
    T = TypeVar('T', int, float)

    @dataclass
    class A(Generic[T]):
        value: T

    def func1(x: T) -> A[T]:
        return A[T](x)

    def func2(x: T) -> A[T]:
        return A[T](x + x)

    pl = sl.Pipeline()
    pl[int] = 1
    pl.insert(func1)
    pl.insert(func2)
    assert pl.compute(A[int]) == A[int](2)


def test_insert_can_replace_generic_param_with_generic_provider() -> None:
    T = TypeVar('T', int, float)

    @dataclass
    class A(Generic[T]):
        value: T

    def func(x: T) -> A[T]:
        return A[T](x + x)

    pl = sl.Pipeline()
    pl[int] = 1
    pl[A[T]] = A[T](1)  # type: ignore[valid-type]
    assert pl.compute(A[int]) == A[int](1)
    pl.insert(func)
    assert pl.compute(A[int]) == A[int](2)


def test_setitem_can_replace_generic_provider_with_generic_param() -> None:
    T = TypeVar('T', int, float)

    @dataclass
    class A(Generic[T]):
        value: T

    def func(x: T) -> A[T]:
        return A[T](x + x)

    pl = sl.Pipeline()
    pl[int] = 1
    pl.insert(func)
    assert pl.compute(A[int]) == A[int](2)
    pl[A[T]] = A[T](1)  # type: ignore[valid-type]
    assert pl.compute(A[int]) == A[int](1)


def test_setitem_can_replace_generic_param_with_generic_param() -> None:
    T = TypeVar('T')

    @dataclass
    class A(Generic[T]):
        value: T

    pl = sl.Pipeline()
    pl[A[T]] = A[T](1)  # type: ignore[valid-type]
    assert pl.compute(A[int]) == A[int](1)
    pl[A[T]] = A[T](2)  # type: ignore[valid-type]
    assert pl.compute(A[int]) == A[int](2)


def test_init_with_params() -> None:
    pl = sl.Pipeline(params={int: 1, float: 2.0})
    assert pl.compute(int) == 1
    assert pl.compute(float) == 2.0


def test_init_with_providers_and_params() -> None:
    def func(x: int, y: float) -> str:
        return f'{x} {y}'

    pl = sl.Pipeline(providers=[func], params={int: 1, float: 2.0})
    assert pl.compute(str) == "1 2.0"


def test_init_with_sciline_Scope_subclass_param_works() -> None:
    T = TypeVar('T')

    class A(sl.Scope[T, int], int):
        ...

    pl = sl.Pipeline(params={A[float]: A(1), A[str]: A(2)})
    assert pl.compute(A[float]) == A(1)
    assert pl.compute(A[str]) == A(2)


def test_building_graph_with_cycle_succeeds() -> None:
    def f(x: int) -> float:
        return float(x)

    def g(x: float) -> int:
        return int(x)

    pipeline = sl.Pipeline([f, g])
    _ = pipeline.get(int)


def test_computing_graph_with_cycle_raises_CycleError() -> None:
    def f(x: int) -> float:
        return float(x)

    def g(x: float) -> int:
        return int(x)

    pipeline = sl.Pipeline([f, g])
    with pytest.raises(sl.scheduler.CycleError):
        pipeline.compute(int)


def test_get_with_single_key_return_task_graph_that_computes_value() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int, int_float_to_str])
    task = pipeline.get(str)
    assert task.compute() == '3;1.5'


@pytest.mark.parametrize('key_type', [tuple, list, iter])
def test_get_with_key_iterable_return_task_graph_that_computes_dict_of_values(
    key_type: Callable[[Any], Any],
) -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    task = pipeline.get(key_type((float, int)))
    assert task.compute() == {float: 1.5, int: 3}


def test_task_graph_compute_can_override_single_key() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    task = pipeline.get(float)
    assert task.compute(int) == 3


def test_task_graph_compute_can_override_key_tuple() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    task = pipeline.get(float)
    assert task.compute((int, float)) == {int: 3, float: 1.5}


def test_task_graph_compute_raises_if_override_keys_outside_graph() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    task = pipeline.get(int)
    # The pipeline knows how to compute int, but the task graph does not
    # as the task graph is fixed at this point.
    with pytest.raises(KeyError):
        task.compute(float)


def test_get_with_NaiveScheduler() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    task = pipeline.get(float, scheduler=sl.scheduler.NaiveScheduler())
    assert task.compute() == 1.5


def test_compute_with_NaiveScheduler() -> None:
    pipeline = sl.Pipeline([int_to_float, make_int])
    res = pipeline.compute(float, scheduler=sl.scheduler.NaiveScheduler())
    assert res == 1.5


def test_bind_and_call_no_function() -> None:
    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(()) == ()


def test_bind_and_call_function_without_args() -> None:
    def func() -> str:
        return "func"

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(func) == "func"


def test_bind_and_call_function_with_1_arg() -> None:
    def func(i: int) -> int:
        return i * 2

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(func) == 6


def test_bind_and_call_function_with_2_arg2() -> None:
    def func(i: int, f: float) -> float:
        return i + f

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call(func) == 4.5


def test_bind_and_call_overrides_default_args() -> None:
    def func(i: int, f: float = -0.5) -> float:
        return i + f

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call(func) == 4.5


def test_bind_and_call_function_in_iterator() -> None:
    def func(i: int) -> int:
        return i * 2

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(iter((func,))) == (6,)


def test_bind_and_call_dataclass_without_args() -> None:
    @dataclass
    class C:
        ...

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(C) == C()


def test_bind_and_call_dataclass_with_1_arg() -> None:
    @dataclass
    class C:
        i: int

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(C) == C(i=3)


def test_bind_and_call_dataclass_with_2_arg2() -> None:
    @dataclass
    class C:
        i: int
        f: float

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call(C) == C(i=3, f=1.5)


def test_bind_and_call_two_functions() -> None:
    def func1(i: int) -> int:
        return 2 * i

    def func2(f: float) -> float:
        return f + 1

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call((func1, func2)) == (6, 2.5)


def test_bind_and_call_two_functions_in_iterator() -> None:
    def func1(i: int) -> int:
        return 2 * i

    def func2(f: float) -> float:
        return f + 1

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call(iter((func1, func2))) == (6, 2.5)


def test_bind_and_call_function_and_dataclass() -> None:
    def func(i: int) -> int:
        return 2 * i

    @dataclass
    class C:
        i: int
        f: float

    pipeline = sl.Pipeline([make_int, int_to_float])
    assert pipeline.bind_and_call((func, C)) == (6, C(i=3, f=1.5))


def test_bind_and_call_function_without_return_annotation() -> None:
    def func(i: int):  # type: ignore[no-untyped-def]
        return 2 * i

    pipeline = sl.Pipeline([make_int])
    assert pipeline.bind_and_call(func) == 6


def test_bind_and_call_generic_function() -> None:
    T = TypeVar('T')
    A = NewType('A', int)
    B = NewType('B', int)

    class G(sl.Scope[T, int], int):
        ...

    def func(a: G[A]) -> int:
        return -4 * a

    pipeline = sl.Pipeline([], params={G[A]: 3, G[B]: 4})
    assert pipeline.bind_and_call(func) == -12


def test_bind_and_call_function_runs_at_end() -> None:
    calls = []

    def a() -> int:
        calls.append('a')
        return 2

    def b() -> float:
        calls.append('b')
        return 3.1

    def c(_i: int) -> None:
        calls.append('c')

    def d(_f: float) -> None:
        calls.append('d')

    pipeline = sl.Pipeline([a, b])
    pipeline.bind_and_call([c, d])

    assert calls.index('a') in (0, 1)
    assert calls.index('b') in (0, 1)
    assert calls.index('c') in (2, 3)
    assert calls.index('d') in (2, 3)


def test_prioritizes_specialized_provider_over_generic() -> None:
    A = NewType('A', str)
    B = NewType('B', str)
    V = TypeVar('V', A, B)

    class H(sl.Scope[V, str], str):
        pass

    def p1(x: V) -> H[V]:
        return H[V]("Generic")

    def p2(x: B) -> H[B]:
        return H[B]("Special")

    pl = sl.Pipeline([p1, p2], params={A: 'A', B: 'B'})

    assert str(pl.compute(H[A])) == "Generic"
    assert str(pl.compute(H[B])) == "Special"


def test_prioritizes_specialized_provider_over_generic_several_typevars() -> None:
    A = NewType('A', str)
    B = NewType('B', str)
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class C(Generic[T1, T2]):
        first: T1
        second: T2
        third: str

    def p1(x: T1, y: T2) -> C[T1, T2]:
        return C(x, y, 'generic')

    def p2(x: A, y: T2) -> C[A, T2]:
        return C(x, y, 'medium generic')

    def p3(x: T2, y: B) -> C[T2, B]:
        return C(x, y, 'generic medium')

    def p4(x: A, y: B) -> C[A, B]:
        return C(x, y, 'special')

    pl = sl.Pipeline([p1, p2, p3, p4], params={A: A('A'), B: B('B')})

    assert pl.compute(C[B, A]) == C('B', 'A', 'generic')
    assert pl.compute(C[A, A]) == C('A', 'A', 'medium generic')
    assert pl.compute(C[B, B]) == C('B', 'B', 'generic medium')
    assert pl.compute(C[A, B]) == C('A', 'B', 'special')


def test_prioritizes_specialized_provider_raises() -> None:
    A = NewType('A', str)
    B = NewType('B', str)
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')

    @dataclass
    class C(Generic[T1, T2]):
        first: T1
        second: T2

    def p1(x: A, y: T1) -> C[A, T1]:
        return C(x, y)

    def p2(x: T1, y: B) -> C[T1, B]:
        return C(x, y)

    pl = sl.Pipeline([p1, p2], params={A: A('A'), B: B('B')})

    with pytest.raises(sl.AmbiguousProvider):
        pl.compute(C[A, B])

    with pytest.raises(sl.UnsatisfiedRequirement):
        pl.compute(C[B, A])


def test_compute_time_handler_allows_for_building_but_not_computing() -> None:
    def func(x: int) -> float:
        return 0.5 * x

    pipeline = sl.Pipeline([func])
    graph = pipeline.get(float, handler=sl.HandleAsComputeTimeException())
    with pytest.raises(sl.UnsatisfiedRequirement):
        graph.compute()


def test_pipeline_copy_simple() -> None:
    a = sl.Pipeline([int_to_float, make_int])
    b = a.copy()
    assert b.compute(int) == 3
    assert b.compute(float) == 1.5


def test_pipeline_copy_dunder() -> None:
    a = sl.Pipeline([int_to_float, make_int])
    from copy import copy

    b = copy(a)
    assert b.compute(int) == 3
    assert b.compute(float) == 1.5


def test_pipeline_copy_with_params() -> None:
    a = sl.Pipeline([int_to_float], params={int: 99})
    b = a.copy()
    assert b.compute(int) == 99
    assert b.compute(float) == 49.5


def test_pipeline_copy_after_insert() -> None:
    a = sl.Pipeline([int_to_float])
    a.insert(make_int)
    b = a.copy()
    assert b.compute(int) == 3
    assert b.compute(float) == 1.5


def test_pipeline_copy_after_setitem() -> None:
    a = sl.Pipeline([int_to_float])
    a[int] = 99
    b = a.copy()
    assert b.compute(int) == 99
    assert b.compute(float) == 49.5


def test_copy_with_generic_providers() -> None:
    Param = TypeVar('Param')

    class Str(sl.Scope[Param, str], str):
        ...

    def parametrized(x: Param) -> Str[Param]:
        return Str(f'{x}')

    def make_float() -> float:
        return 1.5

    def combine(x: Str[int], y: Str[float]) -> str:
        return f"{x};{y}"

    a = sl.Pipeline([make_int, make_float, combine, parametrized])
    b = a.copy()
    assert b.compute(Str[int]) == Str[int]('3')
    assert b.compute(Str[float]) == Str[float]('1.5')
    assert b.compute(str) == '3;1.5'


def test_pipeline_insert_on_copy_does_not_affect_original() -> None:
    a = sl.Pipeline([int_to_float])
    with pytest.raises(sl.UnsatisfiedRequirement):
        a.compute(int)
    b = a.copy()
    b.insert(make_int)
    assert b.compute(int) == 3
    assert b.compute(float) == 1.5
    with pytest.raises(sl.UnsatisfiedRequirement):
        a.compute(int)


def test_pipeline_insert_on_original_does_not_affect_copy() -> None:
    a = sl.Pipeline([int_to_float])
    b = a.copy()
    a.insert(make_int)
    assert a.compute(int) == 3
    assert a.compute(float) == 1.5
    with pytest.raises(sl.UnsatisfiedRequirement):
        b.compute(int)


def test_pipeline_setitem_on_copy_does_not_affect_original() -> None:
    a = sl.Pipeline([int_to_float])
    with pytest.raises(sl.UnsatisfiedRequirement):
        a.compute(int)
    b = a.copy()
    b[int] = 99
    assert b.compute(int) == 99
    assert b.compute(float) == 49.5
    with pytest.raises(sl.UnsatisfiedRequirement):
        a.compute(int)


def test_pipeline_setitem_on_original_does_not_affect_copy() -> None:
    a = sl.Pipeline([int_to_float])
    b = a.copy()
    a[int] = 99
    assert a.compute(int) == 99
    assert a.compute(float) == 49.5
    with pytest.raises(sl.UnsatisfiedRequirement):
        b.compute(int)


def test_pipeline_with_generics_setitem_on_original_does_not_affect_copy() -> None:
    RunType = TypeVar('RunType')

    class RawData(sl.Scope[RunType, int], int):
        ...

    class SquaredData(sl.Scope[RunType, int], int):
        ...

    Sample = NewType('Sample', int)
    Background = NewType('Background', int)
    Result = NewType('Result', int)

    def square(x: RawData[RunType]) -> SquaredData[RunType]:
        return SquaredData[RunType](x * x)

    def process(
        sample: SquaredData[Sample], background: SquaredData[Background]
    ) -> Result:
        return Result(sample + background)

    a = sl.Pipeline(
        [square, process],
        params={
            RawData[Sample]: 5,
            RawData[Background]: 2,
        },
    )
    assert a.compute(Result) == 29
    b = a.copy()
    assert b.compute(Result) == 29
    a[RawData[Sample]] = 7
    assert a.compute(Result) == 53
    assert b.compute(Result) == 29


def test_pipeline_with_generics_setitem_on_copy_does_not_affect_original() -> None:
    RunType = TypeVar('RunType')

    class RawData(sl.Scope[RunType, int], int):
        ...

    class SquaredData(sl.Scope[RunType, int], int):
        ...

    Sample = NewType('Sample', int)
    Background = NewType('Background', int)
    Result = NewType('Result', int)

    def square(x: RawData[RunType]) -> SquaredData[RunType]:
        return SquaredData[RunType](x * x)

    def process(
        sample: SquaredData[Sample], background: SquaredData[Background]
    ) -> Result:
        return Result(sample + background)

    a = sl.Pipeline(
        [square, process],
        params={
            RawData[Sample]: 5,
            RawData[Background]: 2,
        },
    )
    assert a.compute(Result) == 29
    b = a.copy()
    assert b.compute(Result) == 29
    b[RawData[Sample]] = 7
    assert a.compute(Result) == 29
    assert b.compute(Result) == 53


def test_html_repr() -> None:
    pipeline = sl.Pipeline([make_int], params={float: 5.0})
    assert isinstance(pipeline._repr_html_(), str)


def test_pipeline_keyword_only_argument_and_param(
    scheduler: sl.scheduler.Scheduler,
) -> None:
    def fn_with_kwarg(*, y: int) -> str:
        return f'y = {y}'

    pipeline = sl.Pipeline([fn_with_kwarg], params={int: 5})
    assert pipeline.compute(str, scheduler=scheduler) == 'y = 5'


def test_pipeline_mixed_arguments(scheduler: sl.scheduler.Scheduler) -> None:
    def no_args() -> float:
        return 1.2

    def pos_only(a: float) -> list[float]:
        return [a, a]

    def kwarg_only(*, lst: list[float]) -> int:
        return len(lst)

    def mixed_args(i: int, *, lst: list[float]) -> str:
        return f'i = {i}, lst[0] = {lst[0]}'

    pipeline = sl.Pipeline([no_args, pos_only, kwarg_only, mixed_args])
    assert pipeline.compute(str, scheduler=scheduler) == 'i = 2, lst[0] = 1.2'


def test_pipeline_generic_keyword_only(scheduler: sl.scheduler.Scheduler) -> None:
    T = TypeVar('T')
    A = NewType('A', int)
    B = NewType('B', int)

    class G(sl.Scope[T, int], int):
        ...

    def func(*, a: G[A]) -> int:
        return -4 * a

    pipeline = sl.Pipeline([func], params={G[A]: 3, G[B]: 4})
    assert pipeline.compute(int, scheduler=scheduler) == -12


def test_pipeline_detect_missing_argument_typehint() -> None:
    def f(x) -> int:  # type: ignore[no-untyped-def]
        return x  # type:ignore[no-any-return]

    with pytest.raises(ValueError, match='type-hint'):
        sl.Pipeline([f])


def test_pipeline_detect_missing_return_typehint() -> None:
    def f(x: int):  # type: ignore[no-untyped-def]
        return x

    with pytest.raises(ValueError, match='type-hint'):
        sl.Pipeline([f])


def test_does_not_allow_type_argument_outside_of_constraints_flat() -> None:
    T = TypeVar('T', int, float, str)
    T2 = TypeVar('T2', int, float)

    @dataclass
    class M(Generic[T]):
        value: T

    def p1(value: T2) -> M[T2]:
        return M(value)

    pipeline = sl.Pipeline((p1,))
    pipeline[str] = 'abc'
    pipeline[int] = 123

    pipeline.get(M[int])

    with pytest.raises(sl.handler.UnsatisfiedRequirement):
        pipeline.get(M[str])


def test_does_not_allow_type_argument_outside_of_constraints_nested() -> None:
    T = TypeVar('T', int, float, str)

    @dataclass
    class M(Generic[T]):
        value: T

    S = TypeVar('S', M[int], M[float], M[str])
    S2 = TypeVar('S2', M[int], M[float])

    @dataclass
    class N(Generic[S]):
        value: S

    def p1(value: T) -> M[T]:
        return M(value)

    def p2(value: S2) -> N[S2]:
        return N(value)

    pipeline = sl.Pipeline((p1, p2))
    pipeline[str] = 'abc'
    pipeline[int] = 123

    pipeline.get(N[M[int]])

    with pytest.raises(sl.handler.UnsatisfiedRequirement):
        pipeline.get(N[M[str]])


def test_constraints_nested_multiple_typevars() -> None:
    T = TypeVar('T', int, float, str)
    T2 = TypeVar('T2', int, float)

    @dataclass
    class M(Generic[T]):
        v: T

    S = TypeVar('S', M[int], M[float], M[str])
    S2 = TypeVar('S2', M[int], M[float])

    @dataclass
    class N(Generic[S, T]):
        v1: S
        v2: T

    def p1(v: T) -> M[T]:
        return M(v)

    def p2(v1: S2, v2: T2) -> N[S2, T2]:
        return N(v1, v2)

    pipeline = sl.Pipeline((p1, p2))
    pipeline[str] = 'abc'
    pipeline[int] = 123
    pipeline[float] = 3.14

    pipeline.get(N[M[float], int])
    pipeline.get(N[M[int], int])

    with pytest.raises(sl.handler.UnsatisfiedRequirement):
        pipeline.get(N[M[int], str])
    with pytest.raises(sl.handler.UnsatisfiedRequirement):
        pipeline.get(N[M[str], float])


def test_number_of_type_vars_defines_most_specialized() -> None:
    Green = NewType('Green', str)
    Blue = NewType('Blue', str)
    Color = TypeVar('Color', Green, Blue)

    @dataclass
    class Likes(Generic[Color]):
        color: Color

    Preference = TypeVar('Preference')

    @dataclass
    class Person(Generic[Preference, Color]):
        preference: Preference
        hatcolor: Color
        provided_by: str

    def p(c: Color) -> Likes[Color]:
        return Likes(c)

    def p0(p: Preference, c: Color) -> Person[Preference, Color]:
        return Person(p, c, 'p0')

    def p1(c: Color) -> Person[Likes[Color], Color]:
        return Person(Likes(c), c, 'p1')

    def p2(p: Preference) -> Person[Preference, Green]:
        return Person(p, Green('g'), 'p2')

    pipeline = sl.Pipeline((p, p0, p1, p2))
    pipeline[Blue] = 'b'
    pipeline[Green] = 'g'

    # only provided by p0
    assert pipeline.compute(Person[Likes[Green], Blue]) == Person(
        Likes(Green('g')), Blue('b'), 'p0'
    )
    # provided by p1 and p0 but p1 is preferred because it has fewer typevars
    assert pipeline.compute(Person[Likes[Blue], Blue]) == Person(
        Likes(Blue('b')), Blue('b'), 'p1'
    )
    # provided by p2 and p0 but p2 is preferred because it has fewer typevars
    assert pipeline.compute(Person[Likes[Blue], Green]) == Person(
        Likes(Blue('b')), Green('g'), 'p2'
    )

    with pytest.raises(sl.AmbiguousProvider):
        # provided by p1 and p2 with the same number of typevars
        pipeline.get(Person[Likes[Green], Green])


def test_pipeline_with_decorated_provider() -> None:
    R = TypeVar('R')

    def deco(f: Callable[..., R]) -> Callable[..., R]:
        @functools.wraps(f)
        def impl(*args: Any, **kwargs: Any) -> R:
            return f(*args, **kwargs)

        return impl

    provider = deco(int_to_float)
    pipeline = sl.Pipeline([provider], params={int: 3})
    assert pipeline.compute(float) == 1.5


def test_pipeline_with_decorated_provider_keyword_only_arg() -> None:
    R = TypeVar('R')

    def deco(f: Callable[..., R]) -> Callable[..., R]:
        @functools.wraps(f)
        def impl(*args: Any, **kwargs: Any) -> R:
            return f(*args, **kwargs)

        return impl

    @deco
    def foo(*, k: int) -> float:
        return float(k)

    provider = deco(foo)
    pipeline = sl.Pipeline([provider], params={int: 3})
    assert pipeline.compute(float) == 3.0



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import NewType, TypeVar

import pytest

import sciline as sl
from sciline.task_graph import TaskGraph
from sciline.typing import Graph

A = NewType('A', int)
B = NewType('B', int)
T = TypeVar('T', A, B)


class Str(sl.Scope[T, str], str):
    ...


def to_string(x: T) -> Str[T]:
    return Str[T](str(x))


def repeat(a: A, s: Str[B]) -> list[str]:
    return [s] * a


def as_float(x: int) -> float:
    return 0.5 * x


def make_task_graph() -> Graph:
    pl = sl.Pipeline([as_float], params={int: 1})
    return pl.build(float, handler=sl.HandleAsBuildTimeException())


def test_default_scheduler_is_dask_when_dask_available() -> None:
    _ = pytest.importorskip("dask")
    tg = TaskGraph(graph={}, targets=())
    assert isinstance(tg._scheduler, sl.scheduler.DaskScheduler)


def test_compute_returns_value_when_initialized_with_single_key() -> None:
    graph = make_task_graph()
    tg = TaskGraph(graph=graph, targets=float)
    assert tg.compute() == 0.5


def test_compute_returns_dict_when_initialized_with_key_tuple() -> None:
    graph = make_task_graph()
    assert TaskGraph(graph=graph, targets=(float,)).compute() == {float: 0.5}
    assert TaskGraph(graph=graph, targets=(float, int)).compute() == {
        float: 0.5,
        int: 1,
    }


def test_compute_returns_value_when_provided_with_single_key() -> None:
    graph = make_task_graph()
    tg = TaskGraph(graph=graph, targets=float)
    assert tg.compute(int) == 1


def test_compute_returns_dict_when_provided_with_key_tuple() -> None:
    graph = make_task_graph()
    tg = TaskGraph(graph=graph, targets=float)
    assert tg.compute((int, float)) == {int: 1, float: 0.5}


def test_compute_raises_when_provided_with_key_not_in_graph() -> None:
    graph = make_task_graph()
    tg = TaskGraph(graph=graph, targets=float)
    with pytest.raises(KeyError):
        tg.compute(str)
    with pytest.raises(KeyError):
        tg.compute((str, float))


def test_keys_iter() -> None:
    pl = sl.Pipeline([to_string, repeat], params={A: 3, B: 4})
    tg = pl.get(list[str])
    assert len(list(tg.keys())) == 4  # there are no duplicates
    assert set(tg.keys()) == {A, B, Str[B], list[str]}



from collections import defaultdict
from typing import Any, Callable, DefaultDict, Iterable, TypeVar, Union, get_args

from .typing import Item, Key

T = TypeVar('T')
G = TypeVar('G')


def groupby(f: Callable[[T], G], a: Iterable[T]) -> DefaultDict[G, list[T]]:
    g = defaultdict(lambda: [])
    for e in a:
        g[f(e)].append(e)
    return g


def qualname(obj: Any) -> str:
    return str(
        obj.__qualname__ if hasattr(obj, '__qualname__') else obj.__class__.__qualname__
    )


def keyname(key: Union[Key, TypeVar]) -> str:
    if isinstance(key, TypeVar):
        return str(key)
    if isinstance(key, Item):
        return f'{keyname(key.tp)}({keyname(key.label[0].tp)})'
    args = get_args(key)
    if len(args):
        parameters = ', '.join(map(keyname, args))
        return f'{qualname(key)}[{parameters}]'
    return qualname(key)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from typing import Iterator, Mapping, Type, TypeVar

Key = TypeVar('Key')
Value = TypeVar('Value')


class Series(Mapping[Key, Value]):
    """A series of values with labels (row index) and named row dimension."""

    def __init__(self, row_dim: Type[Key], items: Mapping[Key, Value]) -> None:
        """
        Create a new series.

        Parameters
        ----------
        row_dim:
            The row dimension. This must be a type or a type-alias (not an instance).
        items:
            The items of the series.
        """
        self._row_dim = row_dim
        self._map: Mapping[Key, Value] = items

    @property
    def row_dim(self) -> type:
        """The row dimension of the series."""
        return self._row_dim

    def __contains__(self, item: object) -> bool:
        return item in self._map

    def __iter__(self) -> Iterator[Key]:
        return iter(self._map)

    def __len__(self) -> int:
        return len(self._map)

    def __getitem__(self, key: Key) -> Value:
        return self._map[key]

    def __repr__(self) -> str:
        return f"Series(row_dim={self.row_dim.__name__}, {self._map})"

    def _repr_html_(self) -> str:
        return (
            f"<table><tr><th>{self.row_dim.__name__}</th><th>Value</th></tr>"
            + "".join(
                f"<tr><td>{k}</td><td>{v}</td></tr>" for k, v in self._map.items()
            )
            + "</table>"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Series):
            return NotImplemented
        return self.row_dim == other.row_dim and self._map == other._map



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Generic, Optional, TypeVar

import sciline as sl


def test_can_visualize_graph_with_cycle() -> None:
    def int_to_float(x: int) -> float:
        return float(x)

    def float_to_int(x: float) -> int:
        return int(x)

    pipeline = sl.Pipeline([int_to_float, float_to_int])
    graph = pipeline.get(int)
    graph.visualize()


def test_can_visualize_graph_with_unsatisfied_requirements() -> None:
    def int_to_float(x: int) -> float:
        return float(x)

    pipeline = sl.Pipeline([int_to_float])
    pipeline.visualize(int)


def test_generic_types_formatted_without_prefixes() -> None:
    T = TypeVar('T')

    class A(sl.Scope[T, int], int):
        pass

    class B(Generic[T]):
        pass

    class SubA(A[T]):
        pass

    assert sl.visualize._format_type(A[float]).name == 'A[float]'
    assert sl.visualize._format_type(SubA[float]).name == 'SubA[float]'
    assert sl.visualize._format_type(B[float]).name == 'B[float]'


def test_optional_types_formatted_as_their_content() -> None:
    formatted = sl.visualize._format_type(Optional[float])  # type: ignore[arg-type]
    assert formatted.name == 'float'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""Test Pipeline with postponed evaluation of type hints.

Sciline works with postponed evaluation for globally defined types.
But ``typing.get_type_hints`` cannot handle locally defined types
(i.e., defined in a function).
There is nothing we can do about it, so these tests just check that
everything behaves as expected.
"""
from __future__ import annotations

from typing import NewType, TypeVar

import pytest

import sciline as sl

Int = NewType('Int', int)
Str = NewType('Str', str)
T = TypeVar('T')


class Number(sl.Scope[T, int], int):
    ...


class BigNumber(sl.Scope[T, int], int):
    ...


def test_postponed_annotation_builtin() -> None:
    def to_string(x: int) -> str:
        return str(x)

    pl = sl.Pipeline([to_string], params={int: 6})
    assert pl.compute(str) == '6'


def test_postponed_annotation_new_type() -> None:
    def to_string(x: Int) -> Str:
        return Str(str(x))

    pl = sl.Pipeline([to_string], params={Int: 6})
    assert pl.compute(Str) == '6'


def test_postponed_annotation_local_new_type() -> None:
    LInt = NewType('LInt', int)
    LStr = NewType('LStr', str)

    def to_string(x: LInt) -> LStr:
        return LStr(str(x))

    # typing.get_type_hints cannot find LInt and LStr.
    with pytest.raises(NameError):
        sl.Pipeline([to_string], params={LInt: 6})


def test_postponed_annotation_generic_builtin() -> None:
    def lengths(lst: list[str]) -> list[int]:
        return [len(s) for s in lst]

    pl = sl.Pipeline([lengths], params={list[str]: ['abc', 'de', 'f']})
    assert pl.compute(list[int]) == [3, 2, 1]


def test_postponed_annotation_generic_custom() -> None:
    def double(n: Number[T]) -> BigNumber[T]:
        return BigNumber[T](n * 2)

    pl = sl.Pipeline([double], params={Number[Int]: 3, Number[Str]: 5})
    assert pl.compute(BigNumber[Int]) == BigNumber[Int](6)
    assert pl.compute(BigNumber[Str]) == BigNumber[Str](10)


def test_postponed_annotation_generic_local_custom() -> None:
    LT = TypeVar('LT')

    class LNumber(sl.Scope[LT, int], int):
        ...

    class LBigNumber(sl.Scope[LT, int], int):
        ...

    def double(n: LNumber[LT]) -> LBigNumber[LT]:
        return LBigNumber[LT](n * 2)

    # typing.get_type_hints cannot find LNumber and LBigNumber.
    with pytest.raises(NameError):
        sl.Pipeline([double], params={LNumber[Int]: 3, LNumber[Str]: 5})



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""A Sphinx extension for improved typehints rendering.

`sphinx-autodoc-typehints <https://github.com/tox-dev/sphinx-autodoc-typehints>`_
implements formatting for typehints but often results in hard-to-read annotations
for :class:`typing.NewType`.
As those may be ubiquitous when using Sciline, this extension improves formatting.

Usage
-----

This extension relies on
`sphinx-autodoc-typehints <https://github.com/tox-dev/sphinx-autodoc-typehints>`_
for the core implementation of typehints rendering.
Install it as well as Sciline and add both to ``extensions`` in your conf.py:

.. code-block:: python

    extensions = [
        ...,
        'sphinx-autodoc-typehints',
        'sciline.sphinxext.domain_types'
    ]

``sciline.sphinxext.domain_types`` registers a custom ``typehints_formatter``
used by ``sphinx-autodoc-typehints``, so do not define your own!

Options
~~~~~~~

- ``sciline_domain_types_prefix`` (default ``''``): Strip this prefix from type names.
  For example, if a type is defined as ``mypackage.types.SomeType``, setting
  ``sciline_domain_types_prefix = 'mypackage'`` results in this type being rendered
  as ``types.SomeType``.
- ``sciline_domain_types_aliases`` (default ``{}``):
  This extension can render aliases instead of the regular type names.
  This is useful, e.g., for abbreviating long types or types defined
  in internal modules. For example,

  .. code-block:: python

      sciline_domain_types_aliases = {'mypackage._types.SomeType': 'mypackage.SomeType'}

  Renders occurrences of ``SomeType`` as ``mypackage.types.SomeType`` instead of using
  the fully qualified name.
"""

from typing import Any, Optional

from sphinx.application import Sphinx
from sphinx.config import Config


def setup(app: Sphinx) -> dict[str, Any]:
    """Setup sciline.sphinxext.domain_types."""
    app.add_config_value(
        'sciline_domain_types_prefix', default='', rebuild='env', types=str
    )
    app.add_config_value(
        'sciline_domain_types_aliases',
        default={},
        rebuild='env',
        types=dict[str, str],
    )
    app.config.typehints_formatter = _typehints_formatter
    return {'version': 1, 'parallel_read_safe': True, 'parallel_write_safe': True}


def _retrieve_name(obj: Any) -> str:
    """Retrieve the name of an object.

    ``obj.__name__`` if possible or its string representation.
    """

    try:
        return str(obj.__name__)
    except AttributeError:
        return str(obj)


def _typehints_formatter(annotation: Any, config: Config) -> Optional[str]:
    """Format typehints with improved NewType handling."""
    prefix = config.sciline_domain_types_prefix
    aliases = config.sciline_domain_types_aliases

    if _is_new_type(annotation):
        return _format_new_type(annotation, prefix, aliases)
    if _is_type_alias_type(annotation):
        return _format_type_alias_type(annotation, prefix, aliases)
    return None


def _is_new_type(annotation: Any) -> bool:
    # TODO Switch to isinstance(key, NewType) once our minimum is Python 3.10
    # Note that we cannot pass mypy in Python<3.10 since NewType is not a type.
    return hasattr(annotation, '__supertype__')


def _format_new_type(annotation: Any, prefix: str, aliases: dict[str, str]) -> str:
    return (
        f'{_internal_link(annotation, "class", prefix)}'
        f' ({_link(annotation.__supertype__, "class",aliases)})'
    )


def _is_type_alias_type(annotation: Any) -> bool:
    try:
        from typing import TypeAliasType  # type: ignore[attr-defined]

        return isinstance(annotation, TypeAliasType)
    except ImportError:
        return False  # pre python 3.12


def _format_type_alias_type(
    annotation: Any, prefix: str, aliases: dict[str, str]
) -> str:
    alias = _internal_link(annotation, "class", prefix, annotation.__type_params__)
    value = _link(
        annotation.__value__, "class", aliases, _get_type_args(annotation.__value__)
    )
    return f'{alias} ({value})'


def _get_type_args(ty: Any) -> tuple[type, ...]:
    if (args := getattr(ty, '__args__', None)) is not None:
        return args  # type: ignore[no-any-return]  # e.g. list[int]
    return ty.__type_params__  # type: ignore[no-any-return]


def _internal_link(
    annotation: Any,
    kind: str,
    prefix: str,
    type_params: Optional[tuple[type, ...]] = None,
) -> str:
    annotation_name = _retrieve_name(annotation)
    target = f'{annotation.__module__}.{annotation_name}'
    label = f'{annotation.__module__.removeprefix(prefix+".")}.{annotation_name}'
    if type_params:
        label += f'[{", ".join(_retrieve_name(ty) for ty in type_params)}]'
    return f':{kind}:`{label} <{target}>`'


def _link(
    ty: type,
    kind: str,
    aliases: dict[str, str],
    type_params: Optional[tuple[type, ...]] = None,
) -> str:
    ty_name = _retrieve_name(ty)
    if ty.__module__ == 'builtins':
        target = ty_name
    else:
        target = f'{ty.__module__}.{ty_name}'
    label = aliases.get(target, target)
    if type_params:
        label += f'[{", ".join(_retrieve_name(ty) for ty in type_params)}]'
    return f':{kind}:`{label} <{target}>`'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from typing import NoReturn, Protocol, TypeVar

from ._provider import ArgSpec, Provider
from .typing import Key

T = TypeVar('T')


class UnsatisfiedRequirement(Exception):
    """Raised when a type cannot be provided."""


class ErrorHandler(Protocol):
    """Error handling protocol for pipelines."""

    def handle_unsatisfied_requirement(self, tp: Key, *explanation: str) -> Provider:
        ...


class HandleAsBuildTimeException(ErrorHandler):
    """
    Error handler used by default.

    This will raise an exception when building the graph, which is helpful for
    ensuring that errors are caught early, before starting costly computation.
    """

    def handle_unsatisfied_requirement(self, tp: Key, *explanation: str) -> NoReturn:
        """Raise an exception when a type cannot be provided."""
        raise UnsatisfiedRequirement('No provider found for type', tp, *explanation)


class HandleAsComputeTimeException(ErrorHandler):
    """
    Error handler used for Pipeline.visualize().

    This avoids raising exceptions when building the graph, which would prevent
    visualization. This is helpful when visualizing a graph that is not yet complete.
    """

    def handle_unsatisfied_requirement(self, tp: Key, *explanation: str) -> Provider:
        """Return a function that raises an exception when called."""

        def unsatisfied_sentinel() -> NoReturn:
            raise UnsatisfiedRequirement('No provider found for type', tp, *explanation)

        return Provider(
            func=unsatisfied_sentinel, arg_spec=ArgSpec.null(), kind='unsatisfied'
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from html import escape
from typing import Iterable, List, Tuple, TypeVar, Union

from ._provider import Provider
from .typing import Item, Key
from .utils import groupby, keyname


def _details(summary: str, body: str) -> str:
    return f'''
    <details>
      <summary>{summary}</summary>
      {body}
    </details>
    '''


def _provider_name(
    p: Tuple[Key, Tuple[Union[Key, TypeVar], ...], List[Provider]]
) -> str:
    key, args, _ = p
    if args:
        # This is always the case, but mypy complains
        if hasattr(key, '__getitem__'):
            return escape(keyname(key[args]))
    return escape(keyname(key))


def _provider_source(
    p: Tuple[Key, Tuple[Union[Key, TypeVar], ...], List[Provider]]
) -> str:
    key, _, (v, *rest) = p
    if v.kind == 'table_cell':
        # This is always the case, but mypy complains
        if isinstance(key, Item):
            return escape(
                f'ParamTable({keyname(key.label[0].tp)}, length={len((v, *rest))})'
            )
    if v.kind == 'function':
        return _details(
            escape(v.location.name),
            escape(f'{v.location.module}.{v.location.name}'),
        )
    return ''


def _provider_value(
    p: Tuple[Key, Tuple[Union[Key, TypeVar], ...], List[Provider]]
) -> str:
    _, _, (v, *_) = p
    if v.kind == 'parameter':
        html = escape(str(v.call({}))).strip()
        return _details(f'{html[:30]}...', html) if len(html) > 30 else html
    return ''


def pipeline_html_repr(
    providers: Iterable[Tuple[Key, Tuple[Union[Key, TypeVar], ...], Provider]]
) -> str:
    def associate_table_values(
        p: Tuple[Key, Tuple[Union[Key, TypeVar], ...], Provider]
    ) -> Tuple[Key, Union[type, Tuple[Union[Key, TypeVar], ...]]]:
        key, args, v = p
        if isinstance(key, Item):
            return (key.label[0].tp, key.tp)
        return (key, args)

    providers_collected = (
        (key, args, [value, *(v for _, _, v in rest)])
        for ((key, args, value), *rest) in groupby(
            associate_table_values,
            providers,
        ).values()
    )
    provider_rows = '\n'.join(
        (
            f'''
        <tr>
          <td scope="row">{_provider_name(p)}</td>
          <td scope="row">{_provider_value(p)}</td>
          <td scope="row">{_provider_source(p)}</th>
        </tr>'''
            for p in sorted(
                providers_collected,
                key=_provider_name,
            )
        )
    )
    return f'''
    <div class="pipeline-html-repr">
      <table>
        <thead>
          <tr>
            <th scope="col">Name</th>
            <th scope="col">Value</th>
            <th scope="col">Source</th>
          </tr>
        </thead>
        <tbody>
          {provider_rows}
        </tbody>
      </table>
    </div>
    '''.strip()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from dataclasses import dataclass
from typing import NewType, TypeVar

import numpy as np
import numpy.typing as npt

import sciline as sl


@dataclass
class RawData:
    data: npt.NDArray[np.float64]
    monitor1: float
    monitor2: float


SampleRun = NewType('SampleRun', int)
BackgroundRun = NewType('BackgroundRun', int)
DetectorMask = NewType('DetectorMask', npt.NDArray[np.int64])
DirectBeam = NewType('DirectBeam', npt.NDArray[np.float64])
SolidAngle = NewType('SolidAngle', npt.NDArray[np.float64])

Run = TypeVar('Run')


# TODO Giving the base twice works with mypy, how can we avoid typing it twice?
class Raw(sl.Scope[Run, RawData], RawData):
    ...


class Masked(sl.Scope[Run, npt.NDArray[np.float64]], npt.NDArray[np.float64]):
    ...


class IncidentMonitor(sl.Scope[Run, float], float):
    ...


class TransmissionMonitor(sl.Scope[Run, float], float):
    ...


class TransmissionFraction(sl.Scope[Run, float], float):
    ...


class IofQ(sl.Scope[Run, npt.NDArray[np.float64]], npt.NDArray[np.float64]):
    ...


BackgroundSubtractedIofQ = NewType('BackgroundSubtractedIofQ', npt.NDArray[np.float64])


def incident_monitor(x: Raw[Run]) -> IncidentMonitor[Run]:
    return IncidentMonitor[Run](x.monitor1)


def transmission_monitor(x: Raw[Run]) -> TransmissionMonitor[Run]:
    return TransmissionMonitor[Run](x.monitor2)


def mask_detector(x: Raw[Run], mask: DetectorMask) -> Masked[Run]:
    return Masked[Run](x.data * mask)


def transmission(
    incident: IncidentMonitor[Run], transmission: TransmissionMonitor[Run]
) -> TransmissionFraction[Run]:
    return TransmissionFraction[Run](incident / transmission)


def iofq(
    x: Masked[Run],
    solid_angle: SolidAngle,
    direct_beam: DirectBeam,
    transmission: TransmissionFraction[Run],
) -> IofQ[Run]:
    return IofQ[Run](x / (solid_angle * direct_beam * transmission))


reduction = [incident_monitor, transmission_monitor, mask_detector, transmission, iofq]
params = {
    Raw[SampleRun]: Raw(RawData(data=np.ones(4), monitor1=1.0, monitor2=2.0)),
    Raw[BackgroundRun]: Raw(RawData(data=np.ones(4) * 1.5, monitor1=1.0, monitor2=4.0)),
    DetectorMask: DetectorMask(np.array([1, 1, 0, 1])),
    SolidAngle: SolidAngle(np.array([1.0, 0.5, 0.25, 0.125])),
    DirectBeam: DirectBeam(np.array(1 / 1.5)),
}


def subtract_background(
    sample: IofQ[SampleRun], background: IofQ[BackgroundRun]
) -> BackgroundSubtractedIofQ:
    return BackgroundSubtractedIofQ(sample - background)


def test_reduction_workflow() -> None:
    # See https://github.com/python/mypy/issues/14661
    pipeline = sl.Pipeline(
        [subtract_background] + reduction,  # type: ignore[operator]
        params=params,
    )

    assert np.array_equal(pipeline.compute(IofQ[SampleRun]), [3, 6, 0, 24])
    assert np.array_equal(pipeline.compute(IofQ[BackgroundRun]), [9, 18, 0, 72])
    assert np.array_equal(pipeline.compute(BackgroundSubtractedIofQ), [-6, -12, 0, -48])
